"""   
Usage:
    academic_tracker author_search <config_json_file> [--test --prev_pub=<file-path> --no_GoogleScholar --no_ORCID --no_Crossref --verbose --silent]
    academic_tracker reference_search <config_json_file> <references_file_or_URL> [--test --prev_pub=<file-path> --PMID_reference --MEDLINE_reference --no_Crossref --verbose --silent]
    academic_tracker find_ORCID <config_json_file> [--verbose --silent]
    academic_tracker find_Google_Scholar <config_json_file> [--verbose --silent]
    academic_tracker add_authors <config_json_file> <authors_file> [--verbose --silent]
    academic_tracker tokenize_reference <references_file_or_URL> [--MEDLINE_reference --verbose --silent]
    academic_tracker gen_reports_and_emails_auth <config_json_file> <publication_json_file> [--test --verbose --silent]
    academic_tracker gen_reports_and_emails_ref <config_json_file> <references_file_or_URL> <publication_json_file> [--test --prev_pub=<file-path> --MEDLINE_reference --verbose --silent]
    
Options:
    -h --help                         Show this screen.
    --version                         Show version.
    --verbose                         Print hidden error messages.
    --silent                          Do not print anything to the screen.
    --test                            Generate pubs and email texts, but do not send emails.
    --prev_pub=<file-path>            Filepath to json or csv with publication ids to ignore. Enter "ignore" for the <file_path> to not look for previous publications.json files in tracker directories.
    
Reference Type Options:    
    --PMID_reference                  Indicates that the reference_file is a PMID file and only PubMed info will be returned.
    --MEDLINE_reference               Indicates that the reference_file is a MEDLINE file.

Search Options:
    --no_GoogleScholar                Don't search Google Scholar.
    --no_ORCID                        Don't search ORCID.
    --no_Crossref                     Don't search Crossref.
"""


import re
import os
import datetime
import sys
import copy
import warnings

warnings.filterwarnings("ignore", module="fuzzywuzzy")

import docopt

from . import __version__
from . import fileio
from . import user_input_checking
from . import ref_srch_webio
from . import webio
from . import ref_srch_emails_and_reports
from . import tracker_schema
from . import athr_srch_modularized
from . import ref_srch_modularized
from . import helper_functions


VERBOSE = True
SILENT = False

def main():
    
    args = docopt.docopt(__doc__, version=str("Academic Tracker ") + __version__)
    
    user_input_checking.cli_inputs_check(args)
    
    global VERBOSE
    VERBOSE = args["--verbose"]
    global SILENT
    SILENT = args["--silent"]
    
    if len(sys.argv) > 1 and sys.argv[1] == "author_search":
        author_search(args["<config_json_file>"], args["--no_ORCID"], 
                      args["--no_GoogleScholar"], args["--no_Crossref"], 
                      args["--test"], args["--prev_pub"])
    elif len(sys.argv) > 1 and sys.argv[1] == "reference_search":
        if args["--PMID_reference"]:
            PMID_reference(args["<config_json_file>"], args["<references_file_or_URL>"], args["--test"])
        else:
            reference_search(args["<config_json_file>"], args["<references_file_or_URL>"], 
                             args["--MEDLINE_reference"], args["--no_Crossref"], 
                             args["--test"], args["--prev_pub"])
    elif len(sys.argv) > 1 and sys.argv[1] == "find_ORCID":
        find_ORCID(args["<config_json_file>"])
    elif len(sys.argv) > 1 and sys.argv[1] == "find_Google_Scholar":
        find_Google_Scholar(args["<config_json_file>"])
    elif len(sys.argv) > 1 and sys.argv[1] == "add_authors":
        add_authors(args["<config_json_file>"], args["<authors_file>"])
    elif len(sys.argv) > 1 and sys.argv[1] == "tokenize_reference":
        tokenize_reference(args["<references_file_or_URL>"], args["--MEDLINE_reference"])
    elif len(sys.argv) > 1 and sys.argv[1] == "gen_reports_and_emails_auth":
        gen_reports_and_emails_auth(args["<config_json_file>"], args["<publication_json_file>"], args["--test"])
    elif len(sys.argv) > 1 and sys.argv[1] == "gen_reports_and_emails_ref":
        gen_reports_and_emails_ref(args["<config_json_file>"], args["<references_file_or_URL>"], 
                                   args["<publication_json_file>"], args["--MEDLINE_reference"], 
                                   args["--test"], args["--prev_pub"])
    else:
        print("Unrecognized command")  
        


def author_search(config_json_filepath, no_ORCID, no_GoogleScholar, no_Crossref, test, prev_pub_filepath):
    """Query sources for publications by author.
    
    Reads in the JSON config file, previous publications JSON file, and checks for errors.
    Then searches PubMed, ORCID, Google Scholar, and Crossref for publications for each author.
    Emails are then created and sent to project leaders or individual authors. Emails and
    publications are saved to file and emails are sent depending on the options entered by the user. 
    See the program docstring for options details.
    
    Args:
        config_json_filepath (str): filepath to the configuration JSON.
        no_ORCID (bool): If True search ORCID else don't. Reduces checking on config JSON if True.
        no_GoogleScholar (bool): if True search Google Scholar else don't. Reduces checking on config JSON if True.
        no_Crossref (bool): If True search Crossref else don't. Reduces checking on config JSON if True.
        test (bool): If True save_dir_name is tracker-test instead of tracker- and emails are not sent.
        prev_pub_filepath (str or None): filepath to the publication JSON to read in.
    """
    
    config_dict = athr_srch_modularized.input_reading_and_checking(config_json_filepath, no_ORCID, no_GoogleScholar, no_Crossref)
    
    ## Create an authors_json for each project in the config_dict and update those authors attributes with the project attributes.
    authors_by_project_dict, config_dict = athr_srch_modularized.generate_internal_data_and_check_authors(config_dict)
                
    ## read in previous publications to ignore
    has_previous_pubs, prev_pubs = fileio.read_previous_publications(prev_pub_filepath)
    if has_previous_pubs:
        user_input_checking.prev_pubs_file_check(prev_pubs)
            
    ## Query sources and build publication_dict.
    publication_dict, prev_pubs = athr_srch_modularized.build_publication_dict(config_dict, prev_pubs, no_ORCID, no_GoogleScholar, no_Crossref)            
    
    save_dir_name = athr_srch_modularized.save_and_send_reports_and_emails(authors_by_project_dict, publication_dict, config_dict, test)
    
    ## combine previous and new publications lists and save
    fileio.save_publications_to_file(save_dir_name, publication_dict, prev_pubs)
    
    helper_functions.vprint("Success. Publications, reports, and emails saved in " + save_dir_name)



def reference_search(config_json_filepath, ref_path_or_URL, MEDLINE_reference, no_Crossref, test, prev_pub_filepath):
    """Query PubMed and Crossref for publications matching a reference.
    
    Read in user inputs and check for error, query sources based on inputs, build 
    emails and reports, save emails, reports, and publications.
    
    Args:
        config_json_filepath (str): filepath to the configuration JSON.
        ref_path_or_URL (str): either a filepath to file to tokenize or a URL to tokenize.
        MEDLINE_reference (bool): If True re_path_or_URL is a filepath to a MEDLINE formatted file.
        no_Crossref (bool): If True search Crossref else don't. Reduces checking on config JSON if True.
        test (bool): If True save_dir_name is tracker-test instead of tracker- and emails are not sent.
        prev_pub_filepath (str or None): filepath to the publication JSON to read in.
    """
    
    config_dict, tokenized_citations, has_previous_pubs, prev_pubs = ref_srch_modularized.input_reading_and_checking(config_json_filepath, ref_path_or_URL, MEDLINE_reference, no_Crossref, prev_pub_filepath)       

    publication_dict, tokenized_citations = ref_srch_modularized.build_publication_dict(config_dict, tokenized_citations, no_Crossref)
            
    save_dir_name = ref_srch_modularized.save_and_send_reports_and_emails(config_dict, tokenized_citations, publication_dict, prev_pubs, has_previous_pubs, test)
            
    fileio.save_publications_to_file(save_dir_name, publication_dict, {})
    fileio.save_json_to_file(save_dir_name, "tokenized_reference.json", tokenized_citations)
    
    helper_functions.vprint("Success. Publications and reports saved in " + save_dir_name)



def PMID_reference(config_json_filepath, ref_path_or_URL, test):
    """Query PubMed to create a publications JSON file from a list of PMIDs.
    
    Args:
        config_json_filepath (str): filepath to the configuration JSON.
        ref_path_or_URL (str): either a filepath to file to tokenize or a URL to tokenize.
        test (bool): If True save_dir_name is tracker-test instead of tracker- and emails are not sent.
    """
    
    ## read in config file
    config_dict = fileio.load_json(config_json_filepath)
    
    ## Get inputs from config file and check them for errors.
    user_input_checking.ref_config_file_check(config_dict, True)
    
    ## Check the DOI file extension and call the correct read in function.
    extension = os.path.splitext(ref_path_or_URL)[1][1:].lower()
    if extension == "docx":
        file_contents = fileio.read_text_from_docx(ref_path_or_URL)
    elif extension == "txt":
        file_contents = fileio.read_text_from_txt(ref_path_or_URL)
    elif extension == "json":
        file_contents = fileio.load_json(ref_path_or_URL)
    else:
        helper_functions.vprint("Unknown file type for PMID file.")
        sys.exit()
    
    if not file_contents:
        helper_functions.vprint("Nothing was read from the PMID file. Make sure the file is not empty or is a supported file type.")
        sys.exit()
    
    PMID_list = [line for line in file_contents.split("\n") if line] if type(file_contents) == str else file_contents
    
    user_input_checking.tracker_validate(PMID_list, tracker_schema.PMID_reference_schema)
    
    helper_functions.vprint("Querying PubMed and building the publication list.")
    publication_dict = ref_srch_webio.build_pub_dict_from_PMID(PMID_list, config_dict["PubMed_search"]["PubMed_email"])
    
    ## Build the save directory name.
    if test:
        save_dir_name = "tracker-test-" + re.sub(r"\-| |\:", "", str(datetime.datetime.now())[2:16])
    else:
        save_dir_name = "tracker-" + re.sub(r"\-| |\:", "", str(datetime.datetime.now())[2:16])
    os.mkdir(save_dir_name)
    
    ## combine previous and new publications lists and save
    fileio.save_publications_to_file(save_dir_name, publication_dict, {})
    helper_functions.vprint("Success. Publications saved in " + save_dir_name)
    


def find_ORCID(config_json_filepath):
    """Query ORCID to find ORCID IDs for authors.
    
    Args:
        config_json_filepath (str): filepath to the configuration JSON.
    """
    
    ## read in config file
    config_dict = fileio.load_json(config_json_filepath)
    
    ## Get inputs from config file and check them for errors.
    user_input_checking.tracker_validate(config_dict, tracker_schema.ORCID_schema)
            
    old_authors_json = copy.deepcopy(config_dict["Authors"])
    
    helper_functions.vprint("Searching ORCID for author's ORCID ids.")
    config_dict["Authors"] = webio.search_ORCID_for_ids(config_dict["ORCID_search"]["ORCID_key"], config_dict["ORCID_search"]["ORCID_secret"], config_dict["Authors"])
    
    if old_authors_json != config_dict["Authors"]:
        ## Build the save directory name.
        save_dir_name = "tracker-" + re.sub(r"\-| |\:", "", str(datetime.datetime.now())[2:16])
        os.mkdir(save_dir_name)
    
        fileio.save_json_to_file(save_dir_name, "configuration.json", config_dict)
        helper_functions.vprint("Success! configuration.json saved in " + save_dir_name)
    else:
        helper_functions.vprint("No authors were matched from the ORCID results. No new file saved.")



def find_Google_Scholar(config_json_filepath):
    """Query Google Scholar to find Scholar IDs for authors.
    
    Args:
        config_json_filepath (str): filepath to the configuration JSON.
    """
    
    ## read in config file
    config_dict = fileio.load_json(config_json_filepath)
    
    ## Get inputs from config file and check them for errors.
    user_input_checking.tracker_validate(config_dict, tracker_schema.Authors_schema)
        
    old_authors_json = copy.deepcopy(config_dict["Authors"])
    
    helper_functions.vprint("Searching Google Scholar for author's scholar ids.")
    config_dict["Authors"] = webio.search_Google_Scholar_for_ids(config_dict["Authors"])
    
    if old_authors_json != config_dict["Authors"]:
        ## Build the save directory name.
        save_dir_name = "tracker-" + re.sub(r"\-| |\:", "", str(datetime.datetime.now())[2:16])
        os.mkdir(save_dir_name)
    
        fileio.save_json_to_file(save_dir_name, "configuration.json", config_dict)
        helper_functions.vprint("Success! configuration.json saved in " + save_dir_name)
    else:
        helper_functions.vprint("No authors were matched from the Google Scholar results. No new file saved.")



def add_authors(config_json_filepath, authors_filepath):
    """Add authors from csv to config JSON.
    
    Args:
        config_json_filepath (str): filepath to the configuration JSON.
        authors_filepath (str): filepath to the csv file with authors to add or modify.
    """
    
    ## read in config file
    config_dict = fileio.load_json(config_json_filepath)
    
    ## Get inputs from config file and check them for errors.
    user_input_checking.tracker_validate(config_dict, tracker_schema.Authors_schema)
    
    ## Check the file extension and call the correct read in function.
    extension = os.path.splitext(authors_filepath)[1][1:]
    if extension == "csv":
        df = fileio.read_csv(authors_filepath)
    else:
        helper_functions.vprint("Unknown file type for author file.")
        sys.exit()
            
    
    required_columns = tracker_schema.config_schema["properties"]["Authors"]["additionalProperties"]["required"] + ["author_id"]
    missing_required = [column for column in required_columns if column not in df.columns]
    if missing_required:
        helper_functions.vprint("Error: The following columns are required but are missing:\n" + "\n".join(missing_required))
        sys.exit()
        
    missing_values = [column for column in required_columns if df.loc[:, column].isnull().values.any()]
    if missing_values:
        helper_functions.vprint("Error: The following columns have null values:\n" + "\n".join(missing_values))
        sys.exit()
    
    for column in required_columns:
        df.loc[:, column] = df.loc[:, column].astype(str)
    
    ## Assuming all list types are string lists.
    author_keys = tracker_schema.config_schema["properties"]["Authors"]["additionalProperties"]["properties"]
    list_type_keys = [key for key in author_keys if "type" in author_keys[key] and author_keys[key]["type"] == "array"]
    for key in list_type_keys:
        if key in df.columns:
            df.loc[:, key] = df.loc[:, key].astype(str)
            df.loc[:, key] = df.loc[:, key].apply(lambda x: x.split(","))
    
    df.index = df.loc[:, "author_id"]
    df = df.drop(["author_id"], axis=1)
    authors_dict = df.to_dict("index")
    config_dict["Authors"].update(authors_dict)
    
    save_dir_name = "tracker-" + re.sub(r"\-| |\:", "", str(datetime.datetime.now())[2:16])
    os.mkdir(save_dir_name)

    fileio.save_json_to_file(save_dir_name, "configuration.json", config_dict)
    helper_functions.vprint("Success! configuration.json saved in " + save_dir_name)
    
  
    
def tokenize_reference(ref_path_or_URL, MEDLINE_reference):
    """Tokenize input reference file.
    
    Args:
        ref_path_or_URL (str): either a filepath to file to tokenize or a URL to tokenize.
    """
    
    tokenized_citations = ref_srch_webio.tokenize_reference_input(ref_path_or_URL, MEDLINE_reference)
    
    report_string = ref_srch_emails_and_reports.create_tokenization_report(tokenized_citations)
    
    save_dir_name = "tracker-" + re.sub(r"\-| |\:", "", str(datetime.datetime.now())[2:16])
    os.mkdir(save_dir_name)
        
    fileio.save_string_to_file(save_dir_name, "tokenization_report.txt", report_string)
            
    fileio.save_json_to_file(save_dir_name, "tokenized_reference.json", tokenized_citations)
    
    helper_functions.vprint("Success! Tokenization files saved in " + save_dir_name)
    


def gen_reports_and_emails_auth(config_json_filepath, publication_json_filepath, test):
    """Generate reports and emails for input publications as if author_search was ran.
    
    Args:
        config_json_filepath (str): filepath to the configuration JSON.
        publication_json_filepath (str): filepath to the publication JSON to read in.
        test (bool): If True save_dir_name is tracker-test instead of tracker- and emails are not sent.
    """
    
    config_dict = fileio.load_json(config_json_filepath)
    
    ## Get inputs from config file and check them for errors.
    user_input_checking.tracker_validate(config_dict, tracker_schema.gen_reports_auth_schema)
    
    ## Create an authors_json for each project in the config_dict and update those authors attributes with the project attributes.
    authors_by_project_dict, config_dict = athr_srch_modularized.generate_internal_data_and_check_authors(config_dict)
    
    ## Read in publications.json
    publication_dict = fileio.load_json(publication_json_filepath)
    user_input_checking.prev_pubs_file_check(publication_dict)
                
    
    save_dir_name = athr_srch_modularized.save_and_send_reports_and_emails(authors_by_project_dict, publication_dict, config_dict, test)
    
    helper_functions.vprint("Success! Reports and emails saved in " + save_dir_name)
    
    


def gen_reports_and_emails_ref(config_json_filepath, ref_path_or_URL, publication_json_filepath, MEDLINE_reference, test, prev_pub_filepath):
    """Generate reports and emails for input publications and reference as if reference_search was ran.
    
    Args:
        config_json_filepath (str): filepath to the configuration JSON.
        ref_path_or_URL (str): either a filepath to file to tokenize or a URL to tokenize.
        publication_json_filepath (str): filepath to the publication JSON to read in.
        MEDLINE_reference (bool): If True re_path_or_URL is a filepath to a MEDLINE formatted file.
        test (bool): If True save_dir_name is tracker-test instead of tracker- and emails are not sent.
        prev_pub_filepath (str or None): filepath to the publication JSON to read in.
    """
    
    ## read in config file
    config_dict = fileio.load_json(config_json_filepath)
    
    ## Get inputs from config file and check them for errors.
    user_input_checking.tracker_validate(config_dict, tracker_schema.gen_reports_ref_schema)
    
    if not prev_pub_filepath or prev_pub_filepath.lower() == "ignore":
        prev_pubs = {}
        has_previous_pubs = False
    else:
        prev_pubs = fileio.load_json(prev_pub_filepath)
        has_previous_pubs = True
    
    if has_previous_pubs:
        user_input_checking.prev_pubs_file_check(prev_pubs)
        
    tokenized_citations = ref_srch_webio.tokenize_reference_input(ref_path_or_URL, MEDLINE_reference) 
    ## Read in publications.json
    publication_dict = fileio.load_json(publication_json_filepath)
    user_input_checking.prev_pubs_file_check(publication_dict)
    
    pub_titles_to_keys = {pub["title"]:pub_id  for pub_id, pub in publication_dict.items() if pub["title"]}
    pub_titles = list(pub_titles_to_keys.keys())
    
    matching_key_for_citation = []
    for citation in tokenized_citations:
        if not citation["pub_dict_key"] or not citation["pub_dict_key"] in publication_dict:
            if citation["DOI"] and webio.DOI_URL + citation["DOI"] in publication_dict:
                citation["pub_dict_key"] = webio.DOI_URL + citation["DOI"]
                matching_key_for_citation.append(webio.DOI_URL + citation["DOI"])
            elif citation["PMID"] and citation["PMID"] in publication_dict:
                citation["pub_dict_key"] = citation["PMID"]
                matching_key_for_citation.append(citation["PMID"])
            elif citation["title"]:
                matches = helper_functions.fuzzy_matches_to_list(citation["title"], pub_titles)
                if matches:    
                    citation["pub_dict_key"] = pub_titles_to_keys[pub_titles[matches[0][0]]]
                    matching_key_for_citation.append(pub_titles_to_keys[pub_titles[matches[0][0]]])
        else:
            matching_key_for_citation.append(citation["pub_dict_key"])
            
    for key in list(publication_dict.keys()):
        if not key in matching_key_for_citation:
            del publication_dict[key]
            
    if not publication_dict:
        helper_functions.vprint("Error: No entries in the publication JSON matched any reference in the provided reference.")
        sys.exit()
            
    
    save_dir_name = ref_srch_modularized.save_and_send_reports_and_emails(config_dict, tokenized_citations, publication_dict, prev_pubs, has_previous_pubs, test)
    
    fileio.save_json_to_file(save_dir_name, "tokenized_reference.json", tokenized_citations)
    
    helper_functions.vprint("Success! Reports and emails saved in " + save_dir_name)


               
                
if __name__ == "__main__":
    
    main()



# -*- coding: utf-8 -*-
"""
Citation Parsing
~~~~~~~~~~~~~~~~

Functions for parsing citations.
"""

import re

import bs4

from . import helper_functions


def parse_text_for_citations(text):
    """Parse text line by line and tokenize it.
    
    The function is aware of MLA, APA, Chicago, Harvard, and Vancouver style citations.
    Although the citation styles the function is aware of have standards for citations 
    in reality these standards are not strictly adhered to by the public. Therefore
    the function uses a more heuristic approach.
    
    Args:
        text (str): The text to parse.
        
    Returns:
        parsed_pubs (dict): the citations tokenized in a dictionary matching the tokenized citations JSON schema.  
    """
 
## A known issue with these regexes is that authors with 2nd, 3rd, etc in thier name won't get picked up, but allowing numbers in causes too many false positives.      
    regex_dict = {"MLA":r"([^0-9!@#$%^*()[\]_+=\\|<>:;'\"{}`~/?]+)\s+\"(.*)\"\s+(.*)",
                  "APA":r"([^0-9!@#$%^*()[\]_+=\\|<>:;'\"{}`~/?]+)\s+\(\d\d\d\d\)\.\s+([^\.]+)\.\s+(.*)",
                  "Chicago":r"([^0-9!@#$%^*()[\]_+=\\|<>:;'\"{}`~/?]+)\s+\"(.*)\"\s+(.*)",
                  "Harvard":r"([^0-9!@#$%^*()[\]_+=\\|<>:;'\"{}`~/?]+)\s+\d\d\d\d\.\s+([^\.]+)\.\s+(.*)",
                  "Vancouver":r"([^0-9!@#$%^*()[\]_+=\\|<>:;'\"{}`~/?.]+)\.\s+([^\.]+)\.\s+(.*)"}
    
    tokenize_function_dict = {"MLA":tokenize_MLA_or_Chicago_authors,
                              "APA":tokenize_APA_or_Harvard_authors,
                              "Chicago":tokenize_MLA_or_Chicago_authors,
                              "Harvard":tokenize_APA_or_Harvard_authors,
                              "Vancouver":tokenize_Vancouver_authors}
        
    parsed_pubs = []
    
    lines = text.split("\n")
    for count, line in enumerate(lines):
        for citation_style, regex in regex_dict.items():
            groups = helper_functions.regex_match_return(regex, line)
            if groups:
                authors = groups[0].strip()
                ## Sanity check to make sure we are looking at author names separated by commas and not a sentence with a comman in it.
                ## Assuming names won't be more than 4 words.
                temp_authors = authors.replace(" and ", ",")
                sanity_check = any([len(author.strip().split(" ")) > 4 for author in temp_authors.split(",")])
                if sanity_check:
                    continue
                title = groups[1].strip()
                tail = groups[2].strip()
                
                tokenized_authors = tokenize_function_dict[citation_style](authors)
                
                match = helper_functions.regex_match_return(r"(?i).*pmid:\s*(\d+).*", tail)
                pmid = match[0] if match else None
                
                match = helper_functions.regex_match_return(r"(?i).*doi:\s*([^\s]+\w).*", tail)
                if match:
                    doi = match[0].lower()
                    if "doi.org" in doi:
                        match = helper_functions.regex_match_return(r".*doi.org/(.*)", doi)
                        if match:
                            doi = match[0]
                else:
                    doi = None
                            
                parsed_pubs.append({"authors":tokenized_authors, "title":title, "PMID":pmid, "DOI":doi, "reference_line":line.strip(), "pub_dict_key":""})
                break
            
    return parsed_pubs



def tokenize_Vancouver_authors(authors_string):
    """Tokenize authors based on Vancouver citation style.
    
    Args:
        authors_string (str): string with the authors to tokenize.
        
    Returns:
        (list): list of dictionaries with the authors last names and initials. [{"last":lastname, "initials":initials}, ...]
    """
    
    authors_string = authors_string.replace("...", "")
    authors_string = authors_string.replace("&", ",")
    authors_string = authors_string.replace(" and ", ",")
    authors_string = authors_string.replace("et al", "")
    
    names = authors_string.split(",")
    names = [name.strip() for name in names if name.strip()]
    names = [name.split(" ") for name in names]
    
    return [{"last":name[0], "initials":name[1]} if len(name) >1 else {"last":name[0], "initials":""} for name in names]




def tokenize_MLA_or_Chicago_authors(authors_string):
    """Tokenize authors based on MLA or Chicago citation style.
    
    Args:
        authors_string (str): string with the authors to tokenize.
        
    Returns:
        (list): list of dictionaries with the authors first, middle, and last names. [{"first":firstname, "middle":middlename, "last":lastname}, ...]
    """
    
    authors_string = authors_string.replace("...", "")
    authors_string = authors_string.replace(" and ", ",")
    authors_string = authors_string.replace("&", ",")
    authors_string = authors_string.replace("et al.", "")
    
    authors_string = authors_string.strip()
    
    names = authors_string.split(",")
    names = [name.strip() for name in names if name.strip()]
    
    ## The authors_string could have a period at the end that is not part of an initial.
    last_name = names[-1]
    last_name = last_name.split(" ")[-1]
    if len(last_name) > 2 and not re.match(r"([a-zA-Z]\.)+", last_name) and "." in last_name:
        names[-1] = names[-1][:-1]
    
    authors = []
    
    ##The first author in the list doesn't follow the same rules as the rest.
    first_author = names[0]
    if not " " in first_author:
        last = names.pop(0)
        first = names.pop(0)
        first = first.strip()
        first = first.split(" ")
        if len(first) > 1:
            middle = first[1]
            first = first[0]
        else:
            middle = ""
            first = first[0]
            
        authors.append({"first":first, "middle":middle, "last":last})
    
    for name in names:
        name = name.strip()
        tokens = name.split(" ")
        if len(tokens) == 1:
            first = ""
            middle = ""
            last = tokens[0]
        elif len(tokens) > 2:
            first = tokens[0]
            middle = tokens[1]
            last = tokens[2]
        else:
            first = tokens[0]
            middle = ""
            last = tokens[1]
            
        authors.append({"first":first, "middle":middle, "last":last})
        
    return authors




def tokenize_APA_or_Harvard_authors(authors_string):
    """Tokenize authors based on APA or Harvard citation style.
    
    Args:
        authors_string (str): string with the authors to tokenize.
        
    Returns:
        (list): list of dictionaries with the authors last names and initials. [{"last":lastname, "initials":initials}, ...]
    """
    
    authors_string = authors_string.replace("&", ",")
    authors_string = authors_string.replace(" and ", ",")
    authors_string = authors_string.replace("et al.", "")
    authors_string = authors_string.replace(" ", "")
    authors_string = authors_string.replace("...", "")
    
    names_and_initials = authors_string.split(",")
    names_and_initials = [token.strip() for token in names_and_initials if token.strip()]

    authors = []
    previous_token_type = ""
    for token in names_and_initials:
        if re.match(r".*\..*", token):
            if previous_token_type == "last_name":
                authors[-1]["initials"] = token
            else:
                authors.append({"last":"", "initials":token})
            previous_token_type = "initials"
        else:
            authors.append({"last":token, "initials":""})
            previous_token_type = "last_name"
                
            
    return authors
    


def tokenize_myncbi_citations(html):
    """Tokenize the citations on a MyNCBI HTML page.
    
    Note that authors and title can be missing or empty from the webpage.
    
    Args:
        html (str): the html of the MyNCBI page.
        
    Returns:
        parsed_pubs (dict): the citations tokenized in a dictionary matching the tokenized citations JSON schema. 
    """
    
    soup = bs4.BeautifulSoup(html, "html.parser")
    
    parsed_pubs = []
    
    citations = soup.find_all("div", class_ = "ncbi-docsum")
    for citation in citations:
        
        authors_str = citation.find("span", class_ = "authors")
        authors_str = authors_str.text if authors_str else list(citation.children)[1].text
        
        authors_str = authors_str.strip()
        if authors_str and authors_str[-1] == ".":
            authors_str = authors_str[:-1]
        authors = tokenize_Vancouver_authors(authors_str)
        
        ## Some citations don't use Vancouver, so check to see if all initials are blank and if so try Harvard/APA.
        if all([not author["initials"] for author in authors]):
            authors = tokenize_APA_or_Harvard_authors(authors_str)
        
        ## Look for blank authors and remove them.
        authors = [author for author in authors if any([author_attribute for author_attribute in author.values()])]
        
        ## If there is not a span with the class title then the title should be a 
        ## hyperlink that is the 3rd child. 
        title = citation.find("span", class_ = "title")
        if title:
            title = title.text.strip()
        else:
            children = list(citation.children)
            title = "" if children[2].name == "span" else children[2].text.strip()
            
        doi = citation.find("span", class_ = "doi")
        if doi:
            match = helper_functions.regex_match_return(r"(?i).*doi:\s*([^\s]+\w).*", doi.text)
            doi = match[0].lower() if match else ""
        else:
            doi = ""
            
        pmid = citation.find("span", class_ = "pmid")
        if pmid:
            match = helper_functions.regex_match_return(r"(?i).*pmid:\s*(\d+).*", pmid.text)
            pmid = match[0] if match else ""
        else:
            pmid = ""
        
        parsed_pubs.append({"authors":authors, "title":title, "PMID":pmid, "DOI":doi, "reference_line":citation.text.strip(), "pub_dict_key":""})
        
    return parsed_pubs



def parse_MEDLINE_format(text_string):
    """Tokenize text_string based on it being of the MEDLINE format.
    
    Args:
        text_string (str): The string to tokenize.
        
    Returns:
        parsed_pubs (dict): the citations tokenized in a dictionary matching the tokenized citations JSON schema. 
    """
    
    parsed_pubs = []
    pmid = ""
    doi = ""
    title = ""
    found_title = False
    authors = []
    lines = text_string.split("\n")
    for line in lines:
        if line:
            field = line[0:4]
            if field == "PMID":
                pmid = int(line[6:])
            elif field == "TI  ":
                title = line[6:]
                found_title = True
            elif field == "    " and found_title:
                title += line[6:]
            elif not field == "    " and found_title:
                found_title = False
            elif field == "AU  ":
                last_and_initials = line[6:].split(" ")
                if len(last_and_initials) > 2:
                    if sum(1 for char in last_and_initials[1] if char.isupper()) == 1 and len(last_and_initials[1]) > 1:
                        authors.append({"last":last_and_initials[0] + " " + last_and_initials[1], "initials":last_and_initials[2]})
                    else:
                        authors.append({"last":last_and_initials[0], "initials":last_and_initials[1]})
                elif len(last_and_initials) == 2:
                    authors.append({"last":last_and_initials[0], "initials":last_and_initials[1]})
                else:
                    authors.append({"last":last_and_initials[0], "initials":""})
            elif field == "LID ":
                value = line[6:].lower()
                if "doi" in value:
                    doi = value.split(" ")[0]
            elif field == "AID ":
                value = line[6:].lower()
                if "doi" in value:
                    doi = value.split(" ")[0]
                
        else:
            if pmid or doi or title or authors:
                parsed_pubs.append({"authors":authors, "title":title, "PMID":pmid, "DOI":doi, "reference_line":"", "pub_dict_key":""})
            pmid = ""
            doi = ""
            title = ""
            found_title = False
            authors = []
            
    return parsed_pubs







# -*- coding: utf-8 -*-
"""
Tracker Schema
~~~~~~~~~~~~~~

JSON Schema for the input JSON files.
"""

import copy

## TODO add the file inputs and look into adding file existence checking in jsonschema format checker.
cli_schema = {
 "$schema": "https://json-schema.org/draft/2020-12/schema",
 "title": "Command Line Inputs",
 "description": "Input file that contains information for how the program should run.",
 
 "type": "object",
 "properties": {
         "--prev_pub": {"type":["string", "null"], "minLength":1},
         },
         
}


config_schema = {
 "$schema": "https://json-schema.org/draft/2020-12/schema",
 "title": "Configuration JSON",
 "description": "Input file that contains information for how the program should run.",

 "type": "object",
 "properties": {
         "project_descriptions" : {
             "type": "object",
             "minProperties": 1,
             "additionalProperties": {
                     "type":"object",
                     "properties":{
                             "grants": {"type": "array", "minItems":1, "items": {"type": "string", "minLength": 1}},
                             "cutoff_year": {"type": "integer"},
                             "affiliations": {"type": "array", "minItems":1, "items": {"type": "string", "minLength": 1}},
                             "project_report": {"type": "object",
                                                "properties":{
                                                        "columns": {"type": "object",
                                                                    "minProperties":1,
                                                                    "additionalProperties": {"type": "string", "minLength":1}},
                                                        "sort": {"type": "array", "uniqueItems":True, "items": {"type": "string", "minLength":1}, "minItems":1},
                                                        "separator":{"type":"string", "maxLength":1, "minLength":1},
                                                        "column_order":{"type":"array", "uniqueItems":True, "items": {"type": "string", "minLength":1}, "minItems":1},
                                                        "file_format":{"type":"string", "enum":["csv", "xlsx"]},
                                                        "filename":{"type":"string", "minLength":1},
                                                        "template": {"type": "string", "minLength":1},
                                                        "from_email": {"type": "string", "format": "email"},
                                                        "cc_email": {"type": "array",  "items": {"type": "string", "format": "email"}},
                                                        "to_email": {"type": "array",  "items": {"type": "string", "format": "email"}},
                                                        "email_body": {"type": "string", "minLength":1},
                                                        "email_subject": {"type": "string", "minLength":1},},
                                                "dependentRequired":{
                                                        "from_email": ["email_body", "email_subject"],
                                                        "to_email": ["from_email", "email_body", "email_subject"]}},
                             "collaborator_report": {"type": "object",
                                                     "properties":{
                                                             "columns": {"type": "object",
                                                                         "minProperties":1,
                                                                         "additionalProperties": {"type": "string", "minLength":1}},
                                                             "sort": {"type": "array", "uniqueItems":True, "items": {"type": "string", "minLength":1}, "minItems":1},
                                                             "separator":{"type":"string", "maxLength":1, "minLength":1},
                                                             "column_order":{"type":"array", "uniqueItems":True, "items": {"type": "string", "minLength":1}, "minItems":1},
                                                             "file_format":{"type":"string", "enum":["csv", "xlsx"]},
                                                             "filename":{"type":"string", "minLength":1},
                                                             "template": {"type": "string", "minLength":1},
                                                             "from_email": {"type": "string", "format": "email"},
                                                             "cc_email": {"type": "array",  "items": {"type": "string", "format": "email"}},
                                                             "to_email": {"type": "array",  "items": {"type": "string", "format": "email"}},
                                                             "email_body": {"type": "string", "minLength":1},
                                                             "email_subject": {"type": "string", "minLength":1},},
                                                     "dependentRequired":{
                                                             "from_email": ["email_body", "email_subject"],
                                                             "to_email": ["from_email", "email_body", "email_subject"]},},
                             "authors": {"type": "array", "minItems":1, "items": {"type": "string", "minLength": 1}},
                             },
                             
                     "required": ["grants", "affiliations"]
                     }
            },
             
        "ORCID_search" : {"type":"object",
                          "properties": {
                                  "ORCID_key": {"type": "string", "minLength":1},
                                  "ORCID_secret": {"type": "string", "minLength":1}},
                          "required": ["ORCID_key", "ORCID_secret"]},
        "PubMed_search" : {"type":"object",
                          "properties": {
                                  "PubMed_email": {"type": "string", "format":"email"}},
                          "required":["PubMed_email"]},
        "Crossref_search" : {"type":"object",
                          "properties": {
                                  "mailto_email": {"type": "string", "format":"email"}},
                          "required":["mailto_email"]},
        "summary_report" : {"type": "object",
                          "properties":{
                                  "columns": {"type": "object",
                                              "minProperties":1,
                                              "additionalProperties": {"type": "string", "minLength":1}},
                                  "sort": {"type": "array", "uniqueItems":True, "items": {"type": "string", "minLength":1}, "minItems":1},
                                  "separator":{"type":"string", "maxLength":1, "minLength":1},
                                  "column_order":{"type":"array", "uniqueItems":True, "items": {"type": "string", "minLength":1}, "minItems":1},
                                  "file_format":{"type":"string", "enum":["csv", "xlsx"]},
                                  "filename":{"type":"string", "minLength":1},
                                  "template": {"type": "string", "minLength":1},
                                  "from_email": {"type": "string", "format": "email"},
                                  "cc_email": {"type": "array",  "items": {"type": "string", "format": "email"}},
                                  "to_email": {"type": "array",  "items": {"type": "string", "format": "email"}},
                                  "email_body": {"type": "string", "minLength":1},
                                  "email_subject": {"type": "string", "minLength":1},},
                          "dependentRequired":{
                                  "from_email": ["email_body", "email_subject", "to_email"]}},
        "Authors" :  { "type": "object",
                     "minProperties": 1,
                     "additionalProperties": {
                             "type": "object",
                             "properties":{
                                     "first_name": {"type": "string", "minLength":1},
                                     "last_name":{"type": "string", "minLength":1},
                                     "pubmed_name_search": {"type": "string", "minLength":1},
                                     "email":{"type": "string", "format":"email"},
                                     "ORCID":{"type": "string", "pattern":"^\d{4}-\d{4}-\d{4}-\d{3}[0,1,2,3,4,5,6,7,8,9,X]$"},
                                     "grants": {"type": "array", "minItems":1, "items": {"type": "string", "minLength": 1}},
                                     "cutoff_year": {"type": "integer"},
                                     "affiliations": {"type": "array", "minItems":1, "items": {"type": "string", "minLength": 1}},
                                     "scholar_id": {"type": "string", "minLength":1},
                                     "project_report": {"type": "object",
                                                "properties":{
                                                        "columns": {"type": "object",
                                                                    "minProperties":1,
                                                                    "additionalProperties": {"type": "string", "minLength":1}},
                                                        "sort": {"type": "array", "uniqueItems":True, "items": {"type": "string", "minLength":1}, "minItems":1},
                                                        "separator":{"type":"string", "maxLength":1, "minLength":1},
                                                        "column_order":{"type":"array", "uniqueItems":True, "items": {"type": "string", "minLength":1}, "minItems":1},
                                                        "file_format":{"type":"string", "enum":["csv", "xlsx"]},
                                                        "filename":{"type":"string", "minLength":1},
                                                        "template": {"type": "string", "minLength":1},
                                                        "from_email": {"type": "string", "format": "email"},
                                                        "cc_email": {"type": "array",  "items": {"type": "string", "format": "email"}},
                                                        "to_email": {"type": "array",  "items": {"type": "string", "format": "email"}},
                                                        "email_body": {"type": "string", "minLength":1},
                                                        "email_subject": {"type": "string", "minLength":1},},
                                                "dependentRequired":{
                                                        "from_email": ["email_body", "email_subject"],
                                                        "to_email": ["from_email", "email_body", "email_subject"]}},
                                    "collaborator_report": {"type": "object",
                                                     "properties":{
                                                             "columns": {"type": "object",
                                                                         "minProperties":1,
                                                                         "additionalProperties": {"type": "string", "minLength":1}},
                                                             "sort": {"type": "array", "uniqueItems":True, "items": {"type": "string", "minLength":1}, "minItems":1},
                                                             "separator":{"type":"string", "maxLength":1, "minLength":1},
                                                             "column_order":{"type":"array", "uniqueItems":True, "items": {"type": "string", "minLength":1}, "minItems":1},
                                                             "file_format":{"type":"string", "enum":["csv", "xlsx"]},
                                                             "filename":{"type":"string", "minLength":1},
                                                             "template": {"type": "string", "minLength":1},
                                                             "from_email": {"type": "string", "format": "email"},
                                                             "cc_email": {"type": "array",  "items": {"type": "string", "format": "email"}},
                                                             "to_email": {"type": "array",  "items": {"type": "string", "format": "email"}},
                                                             "email_body": {"type": "string", "minLength":1},
                                                             "email_subject": {"type": "string", "minLength":1},},
                                                     "dependentRequired":{
                                                             "from_email": ["email_body", "email_subject"],
                                                             "to_email": ["from_email", "email_body", "email_subject"]},},
                                     },
                             "required" : ["first_name", "last_name", "pubmed_name_search"]

                             }
                       }
                                  
     },
 "required": ["project_descriptions", "ORCID_search", "PubMed_search", "Crossref_search", "Authors"]
}




ORCID_schema = copy.deepcopy(config_schema)
new_properties = {}
new_properties["ORCID_search"] = ORCID_schema["properties"]["ORCID_search"]
new_properties["Authors"] = ORCID_schema["properties"]["Authors"]
ORCID_schema["properties"] = new_properties
ORCID_schema["required"] = ["ORCID_search", "Authors"]


Authors_schema = copy.deepcopy(config_schema)
new_properties = {}
new_properties["Authors"] = Authors_schema["properties"]["Authors"]
Authors_schema["properties"] = new_properties
Authors_schema["required"] = ["Authors"]


ref_config_schema = copy.deepcopy(config_schema)
del ref_config_schema["properties"]["ORCID_search"]
del ref_config_schema["properties"]["Authors"]
del ref_config_schema["properties"]["project_descriptions"]
ref_config_schema["required"] = ["PubMed_search", "Crossref_search"]

gen_reports_auth_schema = copy.deepcopy(config_schema)
del gen_reports_auth_schema["properties"]["PubMed_search"]
del gen_reports_auth_schema["properties"]["ORCID_search"]
del gen_reports_auth_schema["properties"]["Crossref_search"]
gen_reports_auth_schema["required"] = ["project_descriptions", "Authors"]

gen_reports_ref_schema = copy.deepcopy(ref_config_schema)
del gen_reports_ref_schema["properties"]["PubMed_search"]
del gen_reports_ref_schema["properties"]["Crossref_search"]
del gen_reports_ref_schema["required"]




publications_schema={
 "$schema": "https://json-schema.org/draft/2020-12/schema",
 "title": "Publications JSON",
 "description": "Input file that contains information about publications previously found by Academic Tracker.",
 
 "type": "object",
 "additionalProperties": {
         "type": "object",
         "properties": {
                "abstract": {"type":["string", "null"]},
                "authors": {"type":"array", 
                            "minItems":1, 
                            "items": {"type": "object", 
                                      "properties": {
                                              "affiliation": {"type": ["string", "null"]},
                                              "firstname": {"type": ["string", "null"]},
                                              "initials": {"type": ["string", "null"]},
                                              "lastname": {"type": ["string", "null"]},
                                              "author_id" : {"type": "string"}  # optional, only put in if author detected and validated
                                           },
                                        "required": ["affiliation", "firstname", "lastname", "initials"]
                                        }
                            },
                "conclusions": {"type": ["string", "null"]},
                "copyrights": {"type": ["string", "null"]},
                "doi": {"type": ["string", "null"]},
                "journal": {"type": ["string", "null"]},
                "keywords": {"type": ["array", "null"], "items":{"type": ["string", "null"]}},
                "methods": {"type": ["string", "null"]},
                "publication_date": {"type": "object", 
                                     "properties":{"year": {"type": ["integer", "null"]},
                                                   "month": {"type": ["integer", "null"]},
                                                   "day": {"type": ["integer", "null"]}},
                                     "required":["year", "month", "day"]},
                "pubmed_id": {"type": ["string", "null"]},
                "results": {"type": ["string", "null"]},
                "title": {"type": ["string", "null"]},
                "grants": {"type": ["array", "null"], "items":{"type": ["string", "null"]}},
                "PMCID": {"type": ["string", "null"]},
                },
         "required" : ["abstract", "authors", "conclusions", "copyrights", "doi", "journal", "keywords", "methods", "publication_date", "pubmed_id", "results", "title"]
         }
}



tok_schema = {
 "$schema": "https://json-schema.org/draft/2020-12/schema",
 "title": "Tokenized Citations JSON",
 "description": "Input file that contains the tokenized data of a reference file.",
 
 "type": "array",
 "items": {"type": "object",
           "minItems":1,
           "properties": {"authors": {"type": "array",
                                      "items": {"type": "object",
                                                "properties": {"last": {"type":["string", "null"]},
                                                               "initials": {"type":["string", "null"]},
                                                               "first": {"type":["string", "null"]},
                                                               "middle": {"type":["string", "null"]}},
                                                "required": ["last"]}},
                          "title": {"type":["string", "null"]},
                          "PMID": {"type":["string", "null"]},
                          "DOI": {"type":["string", "null"]},
                          "reference_line": {"type":["string", "null"]},
                          "pub_dict_key": {"type":["string", "null"]}},
           "required": ["authors", "title", "PMID", "DOI", "reference_line", "pub_dict_key"]}
}




PMID_reference_schema = {
 "$schema": "https://json-schema.org/draft/2020-12/schema",
 "title": "PMID Reference JSON",
 "description": "Input file that contains a list of PMIDs.",
 
 "type": "array",
 "items": {"type": "string",
           "minItems":1,
           }
}






# -*- coding: utf-8 -*-
"""
Author Search Modularized
~~~~~~~~~~~~~~~~~~~~~~~~~

Modularized pieces of author_search.
"""
import sys
import re
import datetime
import os

from . import user_input_checking
from . import fileio
from . import helper_functions
from . import athr_srch_webio
from . import athr_srch_emails_and_reports
from . import webio


def input_reading_and_checking(config_json_filepath, no_ORCID, no_GoogleScholar, no_Crossref):
    """Read in inputs from user and do error checking.
    
    Args:
        config_json_filepath (str): filepath to the configuration JSON.
        no_ORCID (bool): If True search ORCID else don't. Reduces checking on config JSON if True.
        no_GoogleScholar (bool): if True search Google Scholar else don't. Reduces checking on config JSON if True.
        no_Crossref (bool): If True search Crossref else don't. Reduces checking on config JSON if True.
        
    Returns:
        config_dict (dict): Matches the Configuration file JSON schema.
    """
    
    ## read in config file
    config_dict = fileio.load_json(config_json_filepath)
    
    ## Get inputs from config file and check them for errors.
    user_input_checking.config_file_check(config_dict, no_ORCID, no_GoogleScholar, no_Crossref)
    
    return config_dict




def generate_internal_data_and_check_authors(config_dict):
    """Create authors_by_project_dict and look for authors without projects.
    
    Args:
        config_dict (dict): Matches the Configuration file JSON schema.
        
    Returns:
        authors_by_project_dict (dict): Keys are project names and values are a dictionary of authors and their attributes.
        config_dict (dict): same as input but with author information updated based on project information.
    """
    
    ## Create an authors_json for each project in the config_dict and update those authors attributes with the project attributes.
    authors_by_project_dict = helper_functions.create_authors_by_project_dict(config_dict)
                
    ## Find minimum cutoff_year, and take the union of affiliations and grants for each author.
    helper_functions.adjust_author_attributes(authors_by_project_dict, config_dict)
                    
    ## Look for authors not in any projects and warn user.
    authors_in_projects = {author for project_attributes in config_dict["project_descriptions"].values() if "authors" in project_attributes for author in project_attributes["authors"]}
    authors_not_in_projects = set(config_dict["Authors"].keys()) - authors_in_projects
    projects_without_authors = [project for project, project_attributes in config_dict["project_descriptions"].items() if not "authors" in project_attributes]
    
    if authors_not_in_projects and projects_without_authors:
        helper_functions.vprint("Warning: The following authors in the Authors section of the configuration JSON file are not in any project.")
        for author in authors_not_in_projects:
            helper_functions.vprint(author)
            
    return authors_by_project_dict, config_dict



def build_publication_dict(config_dict, prev_pubs, no_ORCID, no_GoogleScholar, no_Crossref):
    """Query PubMed, ORCID, Google Scholar, and Crossref for publications for the authors.
    
    Args:
        config_dict (dict): Matches the Configuration file JSON schema.
        prev_pubs (dict): Matches the publication JSON schema. Used to ignore publications when querying.
        no_ORCID (bool): If True search ORCID else don't.
        no_GoogleScholar (bool): if True search Google Scholar else don't.
        no_Crossref (bool): If True search Crossref else don't.
        
    Returns:
        publication_dict (dict): The dictionary matching the publication JSON schema.
        prev_pubs (dict): Same as input, but updated with the new publications found.
    """
    
    ## Get publications from PubMed 
    helper_functions.vprint("Finding author's publications. This could take a while.")
    helper_functions.vprint("Searching PubMed.")
    PubMed_publication_dict = athr_srch_webio.search_PubMed_for_pubs(prev_pubs, config_dict["Authors"], config_dict["PubMed_search"]["PubMed_email"])
    prev_pubs.update(PubMed_publication_dict)
    if not no_ORCID:
        helper_functions.vprint("Searching ORCID.")
        ORCID_publication_dict = athr_srch_webio.search_ORCID_for_pubs(prev_pubs, config_dict["ORCID_search"]["ORCID_key"], config_dict["ORCID_search"]["ORCID_secret"], config_dict["Authors"])
        prev_pubs.update(ORCID_publication_dict)
    if not no_GoogleScholar:
        helper_functions.vprint("Searching Google Scholar.")
        Google_Scholar_publication_dict = athr_srch_webio.search_Google_Scholar_for_pubs(prev_pubs, config_dict["Authors"], config_dict["Crossref_search"]["mailto_email"])
        prev_pubs.update(Google_Scholar_publication_dict)
    if not no_Crossref:
        helper_functions.vprint("Searching Crossref.")
        Crossref_publication_dict = athr_srch_webio.search_Crossref_for_pubs(prev_pubs, config_dict["Authors"], config_dict["Crossref_search"]["mailto_email"])
        prev_pubs.update(Crossref_publication_dict)
    
    publication_dict = PubMed_publication_dict
    if not no_ORCID:
        for key, value in ORCID_publication_dict.items():
            if not key in publication_dict:
                publication_dict[key] = value
    if not no_GoogleScholar:
        for key, value in Google_Scholar_publication_dict.items():
            if not key in publication_dict:
                publication_dict[key] = value
    if not no_Crossref:
        for key, value in Crossref_publication_dict.items():
            if not key in publication_dict:
                publication_dict[key] = value
        
        
    if len(publication_dict) == 0:
        helper_functions.vprint("No new publications found.")
        sys.exit()
        
    return publication_dict, prev_pubs



def save_and_send_reports_and_emails(authors_by_project_dict, publication_dict, config_dict, test):
    """Build the summary report and project reports and email them.
    
    Args:
        authors_by_project_dict (dict): Keys are project names and values are a dictionary of authors and their attributes.
        publication_dict (dict): The dictionary matching the publication JSON schema.
        config_dict (dict): Matches the Configuration file JSON schema.
        test (bool): If True save_dir_name is tracker-test instead of tracker- and emails are not sent.
        
    Returns:
        save_dir_name (str): Name of the directory where the emails and reports were saved.
    """
    
    ## Build the save directory name.
    if test:
        save_dir_name = "tracker-test-" + re.sub(r"\-| |\:", "", str(datetime.datetime.now())[2:16])
    else:
        save_dir_name = "tracker-" + re.sub(r"\-| |\:", "", str(datetime.datetime.now())[2:16])
    os.mkdir(save_dir_name)
        
    
    email_messages = athr_srch_emails_and_reports.create_project_reports_and_emails(authors_by_project_dict, publication_dict, config_dict, save_dir_name)
    email_messages["emails"] = email_messages["emails"] + athr_srch_emails_and_reports.create_collaborators_reports_and_emails(publication_dict, config_dict, save_dir_name)["emails"]
            
    if "summary_report" in config_dict:
        
        if "columns" in config_dict["summary_report"]:
            summary_report, summary_filename = athr_srch_emails_and_reports.create_tabular_summary_report(publication_dict, config_dict, authors_by_project_dict, save_dir_name)
        
        else:
            if "template" in config_dict["summary_report"]:
                template = config_dict["summary_report"]["template"]
            else:
                template = athr_srch_emails_and_reports.DEFAULT_SUMMARY_TEMPLATE
                
            if "filename" in config_dict["summary_report"]:
                summary_filename = config_dict["summary_report"]["filename"]
            else:
                summary_filename = "summary_report.txt"
            
            summary_report = athr_srch_emails_and_reports.create_summary_report(publication_dict, config_dict, authors_by_project_dict, template)
            fileio.save_string_to_file(save_dir_name, summary_filename, summary_report)
        
        if "from_email" in config_dict["summary_report"]:
            email_messages["emails"].append({"to":",".join([email for email in config_dict["summary_report"]["to_email"]]),
                                             "from":config_dict["summary_report"]["from_email"],
                                             "cc":",".join([email for email in config_dict["summary_report"]["cc_email"]]) if "cc_email" in config_dict["summary_report"] else "",
                                             "subject":config_dict["summary_report"]["email_subject"],
                                             "body":config_dict["summary_report"]["email_body"],
                                             "attachment":summary_report,
                                             "attachment_filename": summary_filename})
            
    if email_messages["emails"]:
        ## save email messages to file
        fileio.save_emails_to_file(email_messages, save_dir_name)
    
        ## send emails
        if not test:
            webio.send_emails(email_messages)
    
    return save_dir_name





"""

This package has the following modules:
    
``user_input_checking``
    This module contains functions for checking user input.

``tracker_schema``
    This module contains the schema for validating user input.
    
``athr_srch_modularized``
    This module contains functions to complete the author_search command modularized into pieces.
    
``athr_srch_webio``
    This module contains functions for author_search to interface with the internet.
    
``athr_srch_emails_and_reports``
    This module contains functions to create emails and reports for author_search.
    
``ref_srch_modularized``
    This module contains functions to complete the reference_search command modularized into pieces.
    
``ref_srch_webio``
    This module contains functions for reference_search to interface with the internet.
    
``ref_srch_emails_and_reports``
    This module contains functions to create emails and reports for reference_search.
    
``citation_parsing``
    This module contains functions for parsing references and citations for reference_search.

``fileio``
    This module contains functions for reading and writing files.
    
``helper_functions``
    This module contains functions that help the other modules function. The functions do things such as fuzzy matching, regex searching, and printing.
        
``webio``
    This module contains general functions for interfacing with the internet.

"""
__version__ = "1.0.2"



# -*- coding: utf-8 -*-
"""
Author Search Emails and Reports
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Functions to create emails and reports for author_search.
"""

import datetime
import re
import os
import copy

import pandas

from . import helper_functions
from . import fileio

DEFAULT_SUMMARY_TEMPLATE = "<project_loop><project_name>\n<author_loop>\t<author_first> <author_last>:<pub_loop>\n\t\tTitle: <title> \n\t\tAuthors: <authors> \n\t\tJournal: <journal> \n\t\tDOI: <DOI> \n\t\tPMID: <PMID> \n\t\tPMCID: <PMCID> \n\t\tGrants: <grants>\n</pub_loop>\n</author_loop></project_loop>"
DEFAULT_PROJECT_TEMPLATE = "<author_loop><author_first> <author_last>:<pub_loop>\n\tTitle: <title> \n\tAuthors: <authors> \n\tJournal: <journal> \n\tDOI: <DOI> \n\tPMID: <PMID> \n\tPMCID: <PMCID> \n\tGrants: <grants>\n</pub_loop>\n</author_loop>"
DEFAULT_AUTHOR_TEMPLATE = "<author_loop><author_first> <author_last>:<pub_loop>\n\tTitle: <title> \n\tAuthors: <authors> \n\tJournal: <journal> \n\tDOI: <DOI> \n\tPMID: <PMID> \n\tPMCID: <PMCID> \n\tGrants: <grants>\n</pub_loop>\n</author_loop>"


def create_pubs_by_author_dict(publication_dict):
    """Create a dictionary with authors as the keys and values as the pub_ids and grants
    
    Organizes the publication information in an author focused way so other operations are easier.
    
    Args:
        publication_dict (dict): keys and values match the publications JSON file.
        
    Returns:
        pubs_by_author_dict (dict): dictionary where the keys are authors and the values are a dictionary of pub_ids with thier associated grants.
    """
    
    pubs_by_author_dict = {}
    for pub_id, pub_attributes in publication_dict.items():
        for author_attributes in pub_attributes["authors"]:
            if "author_id" in author_attributes:
                author_id = author_attributes["author_id"]
                if author_id in pubs_by_author_dict:
                    pubs_by_author_dict[author_id][pub_id] = pub_attributes["grants"]
                else:
                    pubs_by_author_dict[author_id] = {pub_id : pub_attributes["grants"]}
                    
    return pubs_by_author_dict



def create_project_reports_and_emails(authors_by_project_dict, publication_dict, config_dict, save_dir_name):
    """Create project reports and emails for each project.
    
    For each project in config_dict create a report and optional email.
    Reports are saved in save_dir_name as they are created.
    
    Args:
        authors_by_project_dict (dict): keys are project names from the config file and values are pulled from config_dict["Authors"].
        publication_dict (dict): keys and values match the publications JSON file.
        config_dict (dict): keys and values match the project tracking configuration JSON file.
        save_dir_name (str): directory to save the reports in.
        
    Returns:
        email_messages (dict): keys and values match the email JSON file.
    """
    
    # dict for email messages.
    email_messages = {"creation_date" : str(datetime.datetime.now())[0:16]}
    email_messages["emails"] = []
    
    pubs_by_author_dict = create_pubs_by_author_dict(publication_dict)
    
    for project, project_attributes in config_dict["project_descriptions"].items():
        
        if "project_report" in project_attributes:
            report_attributes = project_attributes["project_report"]
        else:
            continue
        
        ## If to_email is in project then send one email with all authors for the project.
        if "to_email" in report_attributes or not "from_email" in report_attributes:
            
            if "columns" in report_attributes:
                
                file_format = report_attributes["file_format"] if "file_format" in report_attributes else "csv"
                    
                if "filename" in report_attributes:
                    filename = report_attributes["filename"]
                else:
                    if file_format == "csv":
                        filename = project + "_project_report.csv"
                    else:
                        filename = project + "_project_report.xlsx"
                
                report, filename = create_tabular_project_report(publication_dict, config_dict, authors_by_project_dict, pubs_by_author_dict, project, report_attributes, save_dir_name, filename)
            
            else:
                
                template = report_attributes["template"] if "template" in report_attributes else DEFAULT_PROJECT_TEMPLATE
                
                filename = report_attributes["filename"] if "filename" in report_attributes else project + "_project_report.txt"
                
                report = create_project_report(publication_dict, config_dict, authors_by_project_dict, project, template)
                fileio.save_string_to_file(save_dir_name, filename, report)
            
            if "from_email" in report_attributes:
                email_messages["emails"].append({"body":report_attributes["email_body"],
                                                 "subject":report_attributes["email_subject"],
                                                 "from":report_attributes["from_email"],
                                                 "to":",".join([email for email in report_attributes["to_email"]]),
                                                 "cc":",".join([email for email in report_attributes["cc_email"]]) if "cc_email" in report_attributes else "",
                                                 "attachment":report,
                                                 "attachment_filename": filename})
            
        else:
            ## If authors is in project send an email to each author in the project.
            ## If neither authors nor to_email is in the project then send emails to all authors that have publications.  
            authors = project_attributes["authors"] if "authors" in project_attributes else pubs_by_author_dict
                
            for author in authors:
                report_attributes = authors_by_project_dict[project][author]["project_report"]
                
                if not author in pubs_by_author_dict:
                    continue
                
                if "columns" in report_attributes:
                    
                    file_format = report_attributes["file_format"] if "file_format" in report_attributes else "csv"
                        
                    if "filename" in report_attributes:
                        filename = report_attributes["filename"]
                    else:
                        if file_format == "csv":
                            filename = project + "_" + author + "_project_report.csv"
                        else:
                            filename = project + "_" + author + "_project_report.xlsx"
                    
                    report, filename = create_tabular_project_report(publication_dict, config_dict, {project:{author:authors_by_project_dict[project][author]}}, pubs_by_author_dict, project, report_attributes, save_dir_name, filename)
                
                else:
                    template = report_attributes["template"] if "template" in report_attributes else DEFAULT_AUTHOR_TEMPLATE
                    
                    filename = report_attributes["filename"] if "filename" in report_attributes else project + "_" + author + "_project_report.txt"
                    
                    report = create_project_report(publication_dict, config_dict, {project:{author:authors_by_project_dict[project][author]}}, project, template, config_dict["Authors"][author]["first_name"], config_dict["Authors"][author]["last_name"])
                    fileio.save_string_to_file(save_dir_name, filename, report)
                
                if "from_email" in report_attributes and "email" in authors_by_project_dict[project][author]:
                    email_messages["emails"].append({"body":authors_by_project_dict[project][author]["project_report"]["email_body"],
                                                     "subject":authors_by_project_dict[project][author]["project_report"]["email_subject"],
                                                     "from":authors_by_project_dict[project][author]["project_report"]["from_email"],
                                                     "to":authors_by_project_dict[project][author]["email"],
                                                     "cc":",".join([email for email in authors_by_project_dict[project][author]["project_report"]["cc_email"]]) if "cc_email" in authors_by_project_dict[project][author]["project_report"] else "",
                                                     "attachment": report,
                                                     "attachment_filename": filename,
                                                     "author":author})            
    
    return email_messages



def create_project_report(publication_dict, config_dict, authors_by_project_dict, project_name, template_string=DEFAULT_PROJECT_TEMPLATE, author_first = "", author_last = ""):
    """Create the project report for the project.
    
    The details of creating project reports are outlined in the documentation.
    Use the information in the config_dict, publication_dict, and authors_by_project_dict 
    to fill in the information in the template_string. If author_first is given then 
    it is assumed the report is actually for a single author and not a whole project.
    
    Args:
        publication_dict (dict): keys and values match the publications JSON file.
        config_dict (dict): keys and values match the project tracking configuration JSON file.
        authors_by_project_dict (dict): keys are project names from the config file and values are pulled from config_dict["Authors"].
        project_name (str): The name of the project.
        template_string (str): Template used to create the project report.
        author_first (str): First name of the author. If not "" the report is assumed to be for 1 author.
        author_last (str): Last name of the author.
    
    Returns:
        template_string (str): The template_string with the appropriate tags replaced with relevant information.        
    """
    
    project_authors = build_author_loop(publication_dict, config_dict, authors_by_project_dict, project_name, template_string)
    
    template_string = re.sub(r"(?s)<author_loop>.*</author_loop>", project_authors, template_string)
    if author_first:
        template_string = template_string.replace("<author_first>", author_first)
        template_string = template_string.replace("<author_last>", author_last)
    
    return template_string




def create_summary_report(publication_dict, config_dict, authors_by_project_dict, template_string=DEFAULT_SUMMARY_TEMPLATE):
    """Create the summary report for the run.
    
    The details of creating summary reports are outlined in the documentation.
    Use the information in the config_dict, publication_dict, and authors_by_project_dict 
    to fill in the information in the template_string.
    
    Args:
        publication_dict (dict): keys and values match the publications JSON file.
        config_dict (dict): keys and values match the project tracking configuration JSON file.
        authors_by_project_dict (dict): keys are project names from the config file and values are pulled from config_dict["Authors"].
        template_string (str): Template used to create the project report.
    
    Returns:
        report_string (str): The report built by replacing the appropriate tags in template_string with relevant information.
    """
    
    project_template = helper_functions.regex_group_return(helper_functions.regex_match_return(r"(?s).*<project_loop>(.*)</project_loop>.*", template_string), 0)
    
    report_string = ""
    for project_name in config_dict["project_descriptions"]:
        project_template_copy = project_template
        
        project_authors = build_author_loop(publication_dict, config_dict, authors_by_project_dict, project_name, template_string)
        
        project_template_copy = re.sub(r"(?s)<author_loop>.*</author_loop>", project_authors, project_template_copy)
        project_template_copy = project_template_copy.replace("<project_name>", project_name)
        
        report_string += project_template_copy
    
    report_string = re.sub(r"(?s)<project_loop>.*</project_loop>", report_string, template_string)
        
    return report_string




simple_publication_keywords_map = {"<abstract>":"abstract",
                                    "<conclusions>":"conclusions",
                                    "<copyrights>":"copyrights",
                                    "<DOI>":"doi",
                                    "<journal>":"journal",
                                    "<keywords>":"keywords",
                                    "<methods>":"methods",
                                    "<PMID>":"pubmed_id",
                                    "<results>":"results",
                                    "<title>":"title",
                                    "<PMCID>":"PMCID",}

pub_authors_keyword_map = {"<pub_author_first>":"firstname",
                           "<pub_author_last>":"lastname",
                           "<pub_author_initials>":"initials",
                           "<pub_author_affiliations>":"affiliation"}

publication_date_keywords_map = {"<publication_year>":"year",
                                 "<publication_month>":"month",
                                 "<publication_day>":"day"}

authors_keywords_map = {"<author_first>":"first_name",
                        "<author_last>":"last_name",
                        "<author_name_search>":"pubmed_name_search",
                        "<author_email>":"email"}

pub_keywords = list(simple_publication_keywords_map.keys()) + list(publication_date_keywords_map.keys()) + ["<first_author>", "<last_author>", "<authors>", "<grants>"]


def build_author_loop(publication_dict, config_dict, authors_by_project_dict, project_name, template_string):
    """Replace tags in template_string with the appropriate information.
    
    Args:
        publication_dict (dict): keys and values match the publications JSON file.
        config_dict (dict): keys and values match the project tracking configuration JSON file.
        authors_by_project_dict (dict): keys are project names from the config file and values are pulled from config_dict["Authors"].
        project_name (str): The name of the project.
        template_string (str): Template used to create the project report.
        
    Returns:
        project_authors (str): The string built by looping over the authors in authors_by_project_dict and using the template_string to build a report.
    """
    
    pubs_by_author_dict = create_pubs_by_author_dict(publication_dict)
    
    author_template = helper_functions.regex_group_return(helper_functions.regex_match_return(r"(?s).*<author_loop>(.*)</author_loop>.*", template_string), 0)
    pub_template = helper_functions.regex_group_return(helper_functions.regex_match_return(r"(?s).*<pub_loop>(.*)</pub_loop>.*", template_string), 0)
    pub_author_template = helper_functions.regex_group_return(helper_functions.regex_match_return(r"(?s).*<pub_author_loop>(.*)</pub_author_loop>.*", template_string), 0)
    
    project_authors = ""
    for author in authors_by_project_dict[project_name]:
        if not author in pubs_by_author_dict:
            continue
        author_template_copy = author_template
        
        authors_pubs = ""
        for pub in pubs_by_author_dict[author]:
            pub_template_copy = pub_template
            
            pub_authors = ""
            for pub_author in publication_dict[pub]["authors"]:
                pub_author_template_copy = pub_author_template
                
                for keyword, pub_author_key in pub_authors_keyword_map.items():
                    pub_author_template_copy = pub_author_template_copy.replace(keyword, str(pub_author[pub_author_key]))
                    
                pub_authors += pub_author_template_copy
                
            pub_template_copy = re.sub(r"(?s)<pub_author_loop>.*</pub_author_loop>", pub_authors, pub_template_copy)
            
            for keyword, pub_key in simple_publication_keywords_map.items():
                pub_template_copy = pub_template_copy.replace(keyword, str(publication_dict[pub][pub_key]))
                
            ## build first and last author
            first_author = str(publication_dict[pub]["authors"][0]["lastname"]) + ", " + str(publication_dict[pub]["authors"][0]["firstname"])
            pub_template_copy = pub_template_copy.replace("<first_author>", first_author)
            
            last_author = str(publication_dict[pub]["authors"][-1]["lastname"]) + ", " + str(publication_dict[pub]["authors"][-1]["firstname"])
            pub_template_copy = pub_template_copy.replace("<last_author>", last_author)
            
            authors = ", ".join([str(author["firstname"]) + " " + str(author["lastname"]) for author in publication_dict[pub]["authors"]])
            pub_template_copy = pub_template_copy.replace("<authors>", authors)
            
            grants = ", ".join(publication_dict[pub]["grants"]) if publication_dict[pub]["grants"] else "None Found"
            pub_template_copy = pub_template_copy.replace("<grants>", grants)
            
            for keyword, date_key in publication_date_keywords_map.items():
                pub_template_copy = pub_template_copy.replace(keyword, str(publication_dict[pub]["publication_date"][date_key]))
                    
            authors_pubs += pub_template_copy
        
        author_template_copy = re.sub(r"(?s)<pub_loop>.*</pub_loop>", authors_pubs, author_template_copy)
        for keyword, auth_key in authors_keywords_map.items():
            author_template_copy = author_template_copy.replace(keyword, str(config_dict["Authors"][author][auth_key]) if auth_key in config_dict["Authors"][author] else "None")
            
        project_authors += author_template_copy
        
    return project_authors




def create_tabular_collaborator_report(publication_dict, config_dict, author, pubs, filename, file_format, save_dir_name):
    """Create a table for a collaborator report and save as either csv or xlsx.
    
    Args:
        publication_dict (dict): keys and values match the publications JSON file.
        config_dict (dict): keys and values match the project tracking configuration JSON file.
        author (str): The key to the author in config_dict["Authors"].
        pubs (dict): Keys are publications for the author and values are the grants associated with that pub.
        filename (str): filename to save the publication under.
        file_format (str): csv or xlsx, determines what format to save in.
        save_dir_name (str): directory to save the report in.
        
    Returns:
        report (str): The text of the report, empty string, or path to the saved xlsx file.
        filename (str): Filename of the report. Made have had an .xlsx added to the end.
    """
    
    default_columns = {"Name":"<pub_author_last>, <pub_author_first>", "Affiliations":"<pub_author_affiliations>"}
    default_sort = ["Name"]
    default_order = ["Name", "Affiliations"]
    
    if "separator" in config_dict["Authors"][author]["collaborator_report"]:
        separator = config_dict["Authors"][author]["collaborator_report"]["separator"]
    else:
        separator = ","
    
    ## Determine whether to build the report to user specifications or use the deaults.
    if "columns" in config_dict["Authors"][author]["collaborator_report"]:
        columns = config_dict["Authors"][author]["collaborator_report"]["columns"]
        
        if "sort" in config_dict["Authors"][author]["collaborator_report"]:
            sort = config_dict["Authors"][author]["collaborator_report"]["sort"]
        else:
            sort = []
            
        if "column_order" in config_dict["Authors"][author]["collaborator_report"]:
            column_order = config_dict["Authors"][author]["collaborator_report"]["column_order"]
        else:
            column_order = list(columns.keys())
    else:
        columns = default_columns
        sort = default_sort
        column_order = default_order
        
    
    
    
    authors_already_added = []
    
    collaborators = []
    for pub in pubs:
        for pub_author in publication_dict[pub]["authors"]:
            
            if ("author_id" in pub_author and pub_author["author_id"] == author) or pub_author in authors_already_added:
                continue
            
            temp_dict = {}
            for column_name, value in columns.items():
                for keyword, pub_author_key in pub_authors_keyword_map.items():
                    value = value.replace(keyword, str(pub_author[pub_author_key]))
                temp_dict[column_name] = value
                
            collaborators.append(temp_dict)
            authors_already_added.append(pub_author)
    
                
    if collaborators:
        df = pandas.DataFrame(collaborators)
        if sort:
            df = df.sort_values(by=sort)
        df = df.drop_duplicates()
        df = df[column_order]
        
        if file_format == "csv":
            report = df.to_csv(index=False, sep=separator, line_terminator="\n")
            fileio.save_string_to_file(save_dir_name, filename, report)
        else:
            ## If the file extension isn't .xlsx then there will be an error, so force it.
            extension = os.path.splitext(filename)[1][1:].lower()
            if not extension == "xlsx":
                filename += ".xlsx"
            
            report = os.path.join(save_dir_name, filename)
            df.to_excel(report, index=False)
    else:
        report = ""
        
    return report, filename
    



def create_collaborator_report(publication_dict, template, author, pubs, filename, save_dir_name):
    """Create a collaborator report from a formatted string.
    
    Loop over all of the author's publications and create a 
    
    Args:
        publication_dict (dict): keys and values match the publications JSON file.
        author (str): The key to the author in config_dict["Authors"].
        pubs (dict): Keys are publications for the author and values are the grants associated with that pub.
        filename (str): filename to save the publication under.
        save_dir_name (str): directory to save the report in.
        
    Returns:
        report (str): The text of the report or an empty string.
    """
    
    authors_already_added = []
    
    pub_author_template = helper_functions.regex_group_return(helper_functions.regex_match_return(r"(?s).*<pub_author_loop>(.*)</pub_author_loop>.*", template), 0)
    
    report = ""
    for pub in pubs:
        for pub_author in publication_dict[pub]["authors"]:
            
            if ("author_id" in pub_author and pub_author["author_id"] == author) or pub_author in authors_already_added:
                continue
            
            pub_author_template_copy = pub_author_template
            
            for keyword, pub_author_key in pub_authors_keyword_map.items():
                pub_author_template_copy = pub_author_template_copy.replace(keyword, str(pub_author[pub_author_key]))
                
            report += pub_author_template_copy
            
            authors_already_added.append(pub_author)
            
    if report:
        report = re.sub(r"(?s)<pub_author_loop>.*</pub_author_loop>", report, template)
        fileio.save_string_to_file(save_dir_name, filename, report)
    
    return report





def create_collaborators_reports_and_emails(publication_dict, config_dict, save_dir_name):
    """Create a report of collaborators for authors in publication_dict.
    
    For each author in publication_dict with an author_id create a csv file with 
    the other authors on their publicaitons.
    
    Args:
        publication_dict (dict): keys and values match the publications JSON file.
        config_dict (dict): keys and values match the project tracking configuration JSON file.
        save_dir_name (str): directory to save the reports in.
        
    Returns:
        email_messages (dict): keys and values match the email JSON file.
    """

    pubs_by_author_dict = create_pubs_by_author_dict(publication_dict)
    
    # dict for email messages.
    email_messages = {"creation_date" : str(datetime.datetime.now())[0:16]}
    email_messages["emails"] = []
    
    for author, pubs in pubs_by_author_dict.items():
        
        ## Skip if the author isn't in the config dict or if the author doesn't have a collaborator report.
        if not author in config_dict["Authors"]:
            continue
        elif not "collaborator_report" in config_dict["Authors"][author]:
            continue
        
        
        if "template" in config_dict["Authors"][author]["collaborator_report"]:
            template = config_dict["Authors"][author]["collaborator_report"]["template"]
            
            if "filename" in config_dict["Authors"][author]["collaborator_report"]:
                filename = config_dict["Authors"][author]["collaborator_report"]["filename"]
            else:
                filename = author + "_collaborators.txt"
                
            report = create_collaborator_report(publication_dict, template, author, pubs, filename, save_dir_name)
        
        else:
        
            if "file_format" in config_dict["Authors"][author]["collaborator_report"]:
                file_format = config_dict["Authors"][author]["collaborator_report"]["file_format"]
            else:
                file_format = "csv"
                
            if "filename" in config_dict["Authors"][author]["collaborator_report"]:
                filename = config_dict["Authors"][author]["collaborator_report"]["filename"]
            else:
                if file_format == "csv":
                    filename = author + "_collaborators.csv"
                else:
                    filename = author + "_collaborators.xlsx"
                    
            report, filename = create_tabular_collaborator_report(publication_dict, config_dict, author, pubs, filename, file_format, save_dir_name)
            
        
        if report:
            if "from_email" in config_dict["Authors"][author]["collaborator_report"]:
                if "to_email" in config_dict["Authors"][author]["collaborator_report"]:
                    to_email = config_dict["Authors"][author]["collaborator_report"]["to_email"]
                else:
                    to_email = config_dict["Authors"][author]["email"] if "email" in config_dict["Authors"][author] else ""
                
                if to_email:
                    email_messages["emails"].append({"body":config_dict["Authors"][author]["collaborator_report"]["email_body"],
                                                     "subject":config_dict["Authors"][author]["collaborator_report"]["email_subject"],
                                                     "from":config_dict["Authors"][author]["collaborator_report"]["from_email"],
                                                     "to":to_email,
                                                     "cc":",".join([email for email in config_dict["Authors"][author]["collaborator_report"]["cc_email"]]) if "cc_email" in config_dict["Authors"][author]["collaborator_report"] else "",
                                                     "attachment": report,
                                                     "attachment_filename": filename,
                                                     "author":author})
        
    return email_messages
        



def create_tabular_summary_report(publication_dict, config_dict, authors_by_project_dict, save_dir_name):
    """Create a pandas DataFrame and save it as Excel or CSV.
    
    Args:
        publication_dict (dict): keys and values match the publications JSON file.
        config_dict (dict): keys and values match the project tracking configuration JSON file.
        authors_by_project_dict (dict): keys are project names from the config file and values are pulled from config_dict["Authors"].
        save_dir_name (str): directory to save the report in.
        
    Returns:
        report (str): Either the text of the report if csv or a relative filepath to where the Excel file is saved.
        filename (str): Filename of the report. Made have had an .xlsx added to the end.
    """
    
    pubs_by_author_dict = create_pubs_by_author_dict(publication_dict)
        
    rows = []
    row_template = copy.deepcopy(config_dict["summary_report"]["columns"])
    
    if "separator" in config_dict["summary_report"]:
        separator = config_dict["summary_report"]["separator"]
    else:
        separator = ","
        
    if "sort" in config_dict["summary_report"]:
        sort = config_dict["summary_report"]["sort"]
    else:
        sort = []
        
    if "column_order" in config_dict["summary_report"]:
        column_order = config_dict["summary_report"]["column_order"]
    else:
        column_order = list(row_template.keys())
        
    if "file_format" in config_dict["summary_report"]:
        file_format = config_dict["summary_report"]["file_format"]
    else:
        file_format = "csv"
        
    if "filename" in config_dict["summary_report"]:
        filename = config_dict["summary_report"]["filename"]
    else:
        filename = "summary_report.csv" if file_format == "csv" else "summary_report.xlsx"
    
    row_string = "".join(row_template.values())  
    has_pub_keywords = True if any([pub_keyword in row_string for pub_keyword in pub_keywords]) else False
    
    if any([pub_author_keyword in row_string for pub_author_keyword in pub_authors_keyword_map.keys()]):
        has_pub_author_keywords = True
    else:
        has_pub_author_keywords = False
    
    
    for project_name, project_attributes in config_dict["project_descriptions"].items():
        
        for author, author_attributes in authors_by_project_dict[project_name].items():
            if not author in pubs_by_author_dict:
                continue
            
            if has_pub_author_keywords or has_pub_keywords:
                for pub in pubs_by_author_dict[author]:
                    
                    if has_pub_author_keywords:
                        for pub_author in publication_dict[pub]["authors"]:
                            
                            rows.append(replace_keywords(row_template, publication_dict, config_dict, project_name, author, pub, pub_author))
                            
                    else:
                        rows.append(replace_keywords(row_template, publication_dict, config_dict, project_name, author, pub))
                        
            else:
                rows.append(replace_keywords(row_template, publication_dict, config_dict, project_name, author))
                
    
    report = ""            
    if rows:
        df = pandas.DataFrame(rows)
        if sort:
            df = df.sort_values(by=sort)
        df = df.drop_duplicates()
        df = df[column_order]
        
        if file_format == "csv":
            report = df.to_csv(index=False, sep=separator, line_terminator="\n")
            fileio.save_string_to_file(save_dir_name, filename, report)
        else:
            ## If the file extension isn't .xlsx then there will be an error, so force it.
            extension = os.path.splitext(filename)[1][1:].lower()
            if not extension == "xlsx":
                filename += ".xlsx"
            
            report = os.path.join(save_dir_name, filename)
            df.to_excel(report, index=False)
            
    return report, filename



def create_tabular_project_report(publication_dict, config_dict, authors_by_project_dict, pubs_by_author_dict, project_name, report_attributes, save_dir_name, filename):
    """Create a pandas DataFrame and save it as Excel or CSV.
    
    Args:
        publication_dict (dict): keys and values match the publications JSON file.
        config_dict (dict): keys and values match the project tracking configuration JSON file.
        authors_by_project_dict (dict): keys are project names from the config file and values are pulled from config_dict["Authors"].
        pubs_by_author_dict (dict): dictionary where the keys are authors and the values are a dictionary of pub_ids with thier associated grants.
        project_name (str): Name of the project.
        report_attributes (dict): Dictionary of the report attributes. Could come from project_descriptions or an author.
        save_dir_name (str): directory to save the report in.
        filename (str): Filename of the report.
        
    Returns:
        report (str): Either the text of the report if csv or a relative filepath to where the Excel file is saved.
        filename (str): Filename of the report. Made have had an .xlsx added to the end.
    """
    
    rows = []
    row_template = copy.deepcopy(report_attributes["columns"])
    
    separator = report_attributes["separator"] if "separator" in report_attributes else ","
    
    sort = report_attributes["sort"] if "sort" in report_attributes else []
    
    column_order = report_attributes["column_order"] if "column_order" in report_attributes else list(row_template.keys())   
    
    file_format = report_attributes["file_format"] if "file_format" in report_attributes else "csv"    
        
    row_string = "".join(row_template.values())
    
    has_pub_keywords = True if any([pub_keyword in row_string for pub_keyword in pub_keywords]) else False
    
    if any([pub_author_keyword in row_string for pub_author_keyword in pub_authors_keyword_map.keys()]):
        has_pub_author_keywords = True
    else:
        has_pub_author_keywords = False
    
    
    for author, author_attributes in authors_by_project_dict[project_name].items():
        if not author in pubs_by_author_dict:
            continue
        
        if has_pub_author_keywords or has_pub_keywords:
            for pub in pubs_by_author_dict[author]:
                
                if has_pub_author_keywords:
                    for pub_author in publication_dict[pub]["authors"]:
                        
                        rows.append(replace_keywords(row_template, publication_dict, config_dict, project_name, author, pub, pub_author))
                        
                else:
                    rows.append(replace_keywords(row_template, publication_dict, config_dict, project_name, author, pub))
                    
        else:
            rows.append(replace_keywords(row_template, publication_dict, config_dict, project_name, author))
                
    
    report = ""            
    if rows:
        df = pandas.DataFrame(rows)
        if sort:
            df = df.sort_values(by=sort)
        df = df.drop_duplicates()
        df = df[column_order]
        
        if file_format == "csv":
            report = df.to_csv(index=False, sep=separator, line_terminator="\n")
            fileio.save_string_to_file(save_dir_name, filename, report)
        else:
            ## If the file extension isn't .xlsx then there will be an error, so force it.
            extension = os.path.splitext(filename)[1][1:].lower()
            if not extension == "xlsx":
                filename += ".xlsx"
            
            report = os.path.join(save_dir_name, filename)
            df.to_excel(report, index=False)
            
    return report, filename
                    
                    
                    

def replace_keywords(template, publication_dict, config_dict, project_name="", author="", pub="", pub_author={}):
    """Replace keywords in the values of the template dictionary.
    
    Args:
        template (dict): keys are column names and values are what the elements of the column should be.
        publication_dict (dict): keys and values match the publications JSON file.
        config_dict (dict): keys and values match the project tracking configuration JSON file.
        project_name (str): the name of the project to replace.
        author (str): the key to the author in config_dict["Authors"].
        pub (str): the key to the pub in publication_dict.
        pub_author (dict): The author in pub.
        
    Returns:
        template_copy (dict): template with the keywords replaced in its values.
    """
    
    template_copy = copy.deepcopy(template)
    
    for key in template_copy:
        
        ## Project keywords
        if project_name:
            template_copy[key] = template_copy[key].replace("<project_name>", project_name)
        
        ## Authors keywords
        if author:
            for keyword, auth_key in authors_keywords_map.items():
                template_copy[key] = template_copy[key].replace(keyword, str(config_dict["Authors"][author][auth_key]) if auth_key in config_dict["Authors"][author] else "None")
        
        ## Publication keywords
        if pub:
            for keyword, pub_key in simple_publication_keywords_map.items():
                template_copy[key] = template_copy[key].replace(keyword, str(publication_dict[pub][pub_key]))
                
            ## build first and last author
            first_author = str(publication_dict[pub]["authors"][0]["lastname"]) + ", " + str(publication_dict[pub]["authors"][0]["firstname"])
            template_copy[key] = template_copy[key].replace("<first_author>", first_author)
            
            last_author = str(publication_dict[pub]["authors"][-1]["lastname"]) + ", " + str(publication_dict[pub]["authors"][-1]["firstname"])
            template_copy[key] = template_copy[key].replace("<last_author>", last_author)
            
            authors = ", ".join([str(author["firstname"]) + " " + str(author["lastname"]) for author in publication_dict[pub]["authors"]])
            template_copy[key] = template_copy[key].replace("<authors>", authors)
            
            grants = ", ".join(publication_dict[pub]["grants"]) if publication_dict[pub]["grants"] else "None Found"
            template_copy[key] = template_copy[key].replace("<grants>", grants)
            
            for keyword, date_key in publication_date_keywords_map.items():
                template_copy[key] = template_copy[key].replace(keyword, str(publication_dict[pub]["publication_date"][date_key]))
        
        ## Pub authors keywords
        if pub_author:
            for keyword, pub_author_key in pub_authors_keyword_map.items():
                template_copy[key] = template_copy[key].replace(keyword, str(pub_author[pub_author_key]))
            
    return template_copy
        





# -*- coding: utf-8 -*-
"""
Helper Functions
~~~~~~~~~~~~~~~~

This module contains helper functions, such as printing, and regex searching.
"""

import re
import copy

import fuzzywuzzy.fuzz

from . import __main__


def vprint(*args, verbosity=0):
    """Print depending on the state of VERBOSE, SILENT, and verbosity.
    
    If the global SILENT is True don't print anything. If verbosity is 0 then 
    print. If verbosity is 1 then VERBOSE must be True to print.
    
    Args:
        verbosity (int): Either 0 or 1 for different levels of verbosity.
    """
    
    VERBOSE = __main__.VERBOSE
    SILENT = __main__.SILENT
    
    if SILENT:
        return
    
    if verbosity == 0:
        print(*args)
    elif VERBOSE and verbosity == 1:
        print(*args)



def create_authors_by_project_dict(config_dict):
    """Create the authors_by_project_dict dict from the config_dict.
    
    Creates a dict where the keys are the projects in the config_dict and the values
    are the authors associated with that project from config_dict["Authors"].
    
    Args:
        config_dict (dict): schema matches the JSON Project Tracking Configuration file.
        
    Returns:
        authors_by_project_dict (dict): keys are the projects in the config_dict and the values are the authors associated with that project from config_dict["Authors"].
    """
    
    authors_by_project_dict = {project:{} for project in config_dict["project_descriptions"]}
    for project, project_attributes in config_dict["project_descriptions"].items():
        if "authors" in project_attributes:
            for author in project_attributes["authors"]:
                if author in config_dict["Authors"]:
                    authors_by_project_dict[project][author] = copy.deepcopy(config_dict["Authors"][author])
                else:
                    vprint("Warning: The author, " + author + ", in the " + project + " project of the project tracking configuration file could not be found in the Authors section of the Configuration JSON file.")
        else:
            authors_by_project_dict[project] = copy.deepcopy(config_dict["Authors"])
    
        for author_attr in authors_by_project_dict[project].values():
            for key in project_attributes:
                author_attr.setdefault(key, project_attributes[key])
                
    return authors_by_project_dict


def adjust_author_attributes(authors_by_project_dict, config_dict):
    """Modifies config_dict with values from authors_by_project_dict
    
    Go through the authors in authors_by_project_dict and find the lowest cutoff_year.
    Also find affiliations and grants and create a union of them across projects.
    Update the authors in config_dict["Authors"].
    
    Args:
        authors_by_project_dict (dict): keys are the projects in the config_dict and the values are the authors associated with that project from config_dict["Authors"].
        config_dict (dict): schema matches the JSON Project Tracking Configuration file.
        
    Returns:
        config_dict (dict): schema matches the JSON Project Tracking Configuration file.
    """
    
    for author, author_attr in config_dict["Authors"].items():
        cutoff_year = author_attr["cutoff_year"] if "cutoff_year" in author_attr else 99999999
        
        affiliations = set(author_attr["affiliations"]) if "affiliations" in author_attr else set()
        
        grants = set(author_attr["grants"]) if "grants" in author_attr else set()
        
        for project, authors_dict in authors_by_project_dict.items():
            if author in authors_dict:
                if "cutoff_year" in authors_dict[author] and authors_dict[author]["cutoff_year"] < cutoff_year:
                    cutoff_year = authors_dict[author]["cutoff_year"]
                    
                if "affiliations" in authors_dict[author]:
                    affiliations |= set(authors_dict[author]["affiliations"])
                    
                if "grants" in authors_dict[author]:
                    grants |= set(authors_dict[author]["grants"])
                    
                if "collaborator_report" in authors_dict[author]:
                    config_dict["Authors"][author]["collaborator_report"] = authors_dict[author]["collaborator_report"]
                    
        affiliations = list(affiliations)
        grants = list(grants)
        grants.sort()
        affiliations.sort()
        config_dict["Authors"][author]["cutoff_year"] = cutoff_year
        config_dict["Authors"][author]["affiliations"] = affiliations
        config_dict["Authors"][author]["grants"] = grants
        
    return config_dict



    
def regex_match_return(regex, string_to_match):
    """Return the groups matched in the regex if the regex matches.
    
    regex is delivered to re.match() with string_to_match, and if there is a match 
    the match.groups() is returned, otherwise an empty tuple is returned.
    
    Args:
        regex (str): A string with a regular expression to be delivered to re.match().
        string_to_match (str): The string to match with the regex.
        
    Returns:
        (tuple): either the tuple of the matched groups in the regex or an empty tuple if a match wasn't found.
    """
    
    match = re.match(regex, string_to_match)
    return match.groups() if match else ()



def regex_group_return(regex_groups, group_index):
    """Return the group in the regex_groups indicated by group_index if it exists, else return empty string.
    
    If group_index is out of range of the regex_groups an empty string is retruned.
    
    Args:
        regex_groups (tuple): A tuple returned from a matched regex.groups() call.
        group_number (int): The index of the regex_groups to return.
        
    Returns:
        (str): Either emtpy string or the group string matched by the regex.
    """
    
    return regex_groups[group_index] if group_index < len(regex_groups) and regex_groups[group_index] else ""



def regex_search_return(regex, string_to_search):
    """Return the groups matched in the regex if the regex matches.
    
    regex is delivered to re.search() with string_to_search, and if there is a match 
    the match.groups() is returned, otherwise an empty tuple is returned.
    
    Args:
        regex (str): A string with a regular expression to be delivered to re.search().
        string_to_search (str): The string to match with the regex.
    
    Returns:
        (tuple): either the tuple of the matched groups in the regex or an empty tuple if a match wasn't found.
    """
    
    match = re.search(regex, string_to_search)
    return match.groups() if match else ()



    
def match_authors_in_pub_PubMed(authors_json, author_list):
    """Look for matching authors in PubMed pub data.
    
    Goes through the author list from PubMed and tries matching to an author in 
    authors_json using firstname, lastname, and affiliations.
    
    Args:
        authors_json (dict): keys are authors and values are author attributes. Matches authors JSON schema.
        author_list (list): list of dicts where each dict is attributes of an author.
        
    Returns:
        author_list (list): either the author list with matched authors containing an additional author_id attribute, or an empty list if no authors were matched.
    """
    
    ## pub.authors are dictionaries with lastname, firstname, initials, and affiliation. firstname can have initials at the end ex "Andrew P"
    publication_has_affiliated_author = False
    for author_items in author_list:
        author_items_affiliation = str(author_items.get("affiliation")).lower()
        author_items_first_name = str(author_items.get("firstname")).lower()
        author_items_last_name = str(author_items.get("lastname")).lower()
        
        for author, author_attributes in authors_json.items():
        
            ## Match the author's first and last name and then match affiliations.
            ## The first name is matched with an additional .* to try and allow for the addition of initials, but this could cause bad matches. 
            ## For example the name Hu will match Hubert. Counting on the last name to reduce errors.
            ## Note that the PubMed query will return publications where the author is just a collaborater, so not finding a match in the authors isn't uncommon.
            if re.match(author_attributes["first_name"].lower() + ".*", author_items_first_name) and author_attributes["last_name"].lower() == author_items_last_name:
                ## affiliations are sets of strings so the intersection should have at least 1 string for a match.
                
                if any([affiliation.lower() in author_items_affiliation for affiliation in author_attributes["affiliations"]]):
                    publication_has_affiliated_author = True
                    author_items["author_id"] = author
                    break
                
    return author_list if publication_has_affiliated_author else []

    
    
    
def match_authors_in_pub_Crossref(authors_json, author_list):
    """Look for matching authors in Crossref pub data.
    
    Goes through the author list from Crossref and tries matching to an author in 
    authors_json using firstname, lastname, and affiliations, or ORCID if ORCID
    is present.
    
    Args:
        authors_json (dict): keys are authors and values are author attributes. Matches authors JSON schema.
        author_list (list): list of dicts where each dict is attributes of an author.
        
    Returns:
        author_list (list): either the author list with matched authors containing an additional author_id attribute, or an empty list if no authors were matched.
    """
    
    ## pub.authors are dictionaries with lastname, firstname, initials, and affiliation. firstname can have initials at the end ex "Andrew P"
    publication_has_affiliated_author = False
    for author_items in author_list:
        ## If the author has no affiliation or ORCID id then go to the next.
        if not author_items["affiliation"] and not "ORCID" in author_items:
            continue
        
        if author_items["affiliation"] and "name" in author_items["affiliation"][0]:
            author_items_affiliation = author_items["affiliation"][0]["name"].lower()
        else:
            author_items_affiliation = []
        
        author_items_first_name = str(author_items.get("given")).lower() if "given" in author_items else ""
        author_items_last_name = str(author_items.get("family")).lower()
        
        if "ORCID" in author_items:
            ORCID_id = regex_search_return(r"(\d{4}-\d{4}-\d{4}-\d{3}[0,1,2,3,4,5,6,7,8,9,X])", author_items["ORCID"])[0]
        else:
            ORCID_id = ""
        
        for author, author_attributes in authors_json.items():
        
            ## Match the author's first and last name and then match affiliations.
            ## The first name is matched with an additional .* to try and allow for the addition of initials, but this could cause bad matches. 
            ## For example the name Hu will match Hubert. Counting on the last name to reduce errors.
            ## Note that the PubMed query will return publications where the author is just a collaborater, so not finding a match in the authors isn't uncommon.
            if re.match(author_attributes["first_name"].lower() + ".*", author_items_first_name) and author_attributes["last_name"].lower() == author_items_last_name:
                ## affiliations are sets of strings so the intersection should have at least 1 string for a match.
                if any([affiliation.lower() in author_items_affiliation for affiliation in author_attributes["affiliations"]]):
                    publication_has_affiliated_author = True
                    author_items["author_id"] = author
                    break
                
            elif ORCID_id and "ORCID" in author_attributes and ORCID_id == author_attributes["ORCID"]:
                publication_has_affiliated_author = True
                author_items["author_id"] = author
                break
                
    if publication_has_affiliated_author:
        return author_list
    else:
        return []




def modify_pub_dict_for_saving(pub):
    """Convert pymed.PubMedArticle to a dictionary and modify it for saving.
    
    Converts a pymed.PubMedArticle to a dictionary, deletes the "xml" key, and 
    converts the "publication_date" key to a string.
    
    Args:
        pub (pymed.PubMedArticle): publication to convert to a dictionary.
        
    Returns:
        pub_dict (dict): pub converted to a dictionary. Keys are "pubmed_id", "title", "abstract", 
        "keywords", "journal", "publication_date", "authors", "methods", "conclusions", "results", "copyrights", and "doi"
    """
    
    pub_dict = pub.toDict()
    
    pub_dict["grants"] = [grant.text for grant in pub.xml.findall(".//GrantID")]
    PMC_id_elements = pub.xml.findall(".//ArticleId[@IdType='pmc']")
    pub_dict["PMCID"] = PMC_id_elements[0].text if PMC_id_elements else None
        
    del pub_dict["xml"]
    pub_dict["publication_date"] = {"year":pub_dict["publication_date"].year, "month":pub_dict["publication_date"].month, "day":pub_dict["publication_date"].day}
    pub_dict["pubmed_id"] = pub_dict["pubmed_id"].split("\n")[0]
    
    return pub_dict





def is_fuzzy_match_to_list(str_to_match, list_to_match):
    """True if string is a 90 or higher ratio match to any string in list, False otherwise.
    
    Args:
        str_to_match (str): string to compare with list
        list_to_match (list): list of strings to compare with str_to_match
        
    Returns:
        (bool): True if str_to_match is a match to any string in list_tp_match, False otherwise.
    """
    str_to_match = str_to_match.lower()
    list_to_match = [list_string.lower() for list_string in list_to_match]
    
    return any([fuzzywuzzy.fuzz.ratio(str_to_match, list_string) >= 90 for list_string in list_to_match])



def fuzzy_matches_to_list(str_to_match, list_to_match):
    """Return strings and indexes for strings with match ratio that is 90 or higher.
    
    Args:
        str_to_match (str): string to compare with list
        list_to_match (list): list of strings to compare with str_to_match
        
    Returns:
        (list): list of matches (tuples) with each element being the string and its index in list_to_match. [(9, "title 1"), ...]
    """
    str_to_match = str_to_match.lower()
    list_to_match = [list_string.lower() for list_string in list_to_match]
    
    return [(index, list_string) for index, list_string in enumerate(list_to_match) if fuzzywuzzy.fuzz.ratio(str_to_match, list_string) >= 90]



def is_pub_in_publication_dict(pub_id, title, publication_dict, titles=[]):
    """True if pub_id is in publication_dict or title is a fuzzy match to titles in titles.
    
    Check whether the pub_id is in publication_dict. If it isn't then see if there is 
    a fuzzy match in titles. If titles is not provided then get a list of titles 
    from publication_dict.
    
    Args:
        pub_id (str): pub_id to check against in publication_dict to see if it already exists.
        title (str): title corresponding to pub_id to check against titles in publication_dict.
        publication_dict (dict): keys are pub_ids and values are pub attributes.
        titles (list): list of strings that should be titles to fuzzy match to title.
        
    Returns:
        (bool): True if the pub_id is in publication_dict or title is fuzzy matched in titles, False otherwise
    """
    
    if pub_id.lower() in publication_dict:
        return True
    
    if not titles:
        titles = [pub_attr["title"] for pub_attr in publication_dict.values() if pub_attr["title"]]
    
    return True if is_fuzzy_match_to_list(title, titles) else False



def find_duplicate_citations(tokenized_citations):
    """Find citations that are duplicates of each other in tokenized_citations.
    
    Citations can be duplicates in 3 ways. Same PMID, same DOI, or similar enough titles.
    The function goes through each citation and looks for matches on these criteria.
    Then the matches are compared to create unique sets. For instance if citation 1
    matches the PMID in citation 2, and citation 2 matches the DOI in citation 3, but 
    citation 1 and 3 don't match a duplicate set containing all 3 is created. The 
    unique duplicate sets are returned as a list of sorted lists.
    
    Args:
        tokenized_citation (list): list of dictionaries where each dictionary is a citation. Matches the tokenized_reference.json schema.
        
    Returns:
        unique_duplicate_sets (list): list of lists where each element is a list of indexes in tokenized_citations that match each other. The list of indexes is sorted in ascending order.
    """
    
    titles_list = [citation["title"] for citation in tokenized_citations if citation["title"]]
    
    pmids = {}
    dois = {}
    titles = {}
    for count, citation in enumerate(tokenized_citations):
        
        if citation["PMID"]:
            if citation["PMID"] in pmids:
                pmids[citation["PMID"]].append(count)
            else:
                pmids[citation["PMID"]] = [count]
                
        if citation["DOI"]:
            doi = citation["DOI"].lower()
            if doi in dois:
                dois[doi].append(count)
            else:
                dois[doi] = [count]
                
        if citation["title"]:
            fuzzy_matches = fuzzy_matches_to_list(citation["title"], titles_list)
            if len(fuzzy_matches) > 1 and fuzzy_matches[0][1] in titles:
                titles[fuzzy_matches[0][1]].append(count)
            else:
                titles[citation["title"]] = [count]
                
    
    matching_pmids = {tuple(indexes) for pmid, indexes in pmids.items() if len(indexes) > 1}
    matching_dois = {tuple(indexes) for doi, indexes in dois.items() if len(indexes) > 1}
    matching_titles = {tuple(indexes) for title, indexes in titles.items() if len(indexes) > 1}
    
    all_matches = matching_pmids | matching_dois | matching_titles
    
    ## Matches need to be matched recursively to find true sets of matches.
    ## For example if 1 and 2 are found to match on title and 2 and 3 are found to match 
    ## on PMID then we need 1 set of (1,2,3) instead of 2 sets of (1,2) (2,3).
    duplicates_dict = {}
    for index_tuple in list(all_matches):
        for index in index_tuple:
            if index in duplicates_dict:
                for index2 in index_tuple:
                    if index2 != index:
                        duplicates_dict[index].add(index2)
            else:
                duplicates_dict[index] = set()
                for index2 in index_tuple:
                    if index2 != index:
                        duplicates_dict[index].add(index2)
                        
    for index, matches in duplicates_dict.items():
        temp_set = matches
        set_before_loop = matches
        while(True):
            for match in matches:
                temp_set = temp_set | duplicates_dict[match]
            
            if set_before_loop == temp_set:
                break
            else:
                set_before_loop = temp_set
                
        duplicates_dict[index] = temp_set
        
    unique_duplicate_sets = {tuple(duplicate_set) for duplicate_set in duplicates_dict.values()}
    unique_duplicate_sets = [list(duplicate_set) for duplicate_set in unique_duplicate_sets]
    for duplicate_set in unique_duplicate_sets:
        duplicate_set.sort()
    
    return unique_duplicate_sets
    
    

def are_citations_in_pub_dict(tokenized_citations, pub_dict):
    """Determine which citations in tokenized_citations are in pub_dict.
    
    For each citation in tokenized_citations see if it is in pub_dict. Will be 
    True for a citation if the PMID matches, DOI matches, or the title is similar 
    enough.
    
    Args:
        tokenized_citation (list): list of dictionaries where each dictionary is a citation. Matches the tokenized_reference.json schema.
        pub_dict (dict): schema matches the publication.json schema.
    
    Returns:
        (list): list of bools, True if the citation at that index is in pub_dict, False otherwise.
    """
    
    pub_titles = [pub["title"] for pub in pub_dict.values() if pub["title"]]
    pub_dois = [pub["doi"].lower() for pub in pub_dict.values() if pub["doi"]]
    pub_pmids = [pub["pubmed_id"] for pub in pub_dict.values() if pub["pubmed_id"]]
    
    return [True if citation["PMID"] in pub_pmids or citation["DOI"].lower() in pub_dois or is_fuzzy_match_to_list(citation["title"], pub_titles) else False for citation in tokenized_citations]





# -*- coding: utf-8 -*-
"""
Author Search Webio
~~~~~~~~~~~~~~~~~~~

Internet interfacing for author_search.
"""


import time
import copy

import pymed
import orcid
import scholarly
import habanero

from . import helper_functions
from . import webio



TOOL = webio.TOOL
DOI_URL = webio.DOI_URL

PUBLICATION_TEMPLATE = webio.PUBLICATION_TEMPLATE


## TODO get with pymed and add grants and pmcid to PubMedArticle class.
def search_PubMed_for_pubs(prev_pubs, authors_json, from_email):
    """Searhes PubMed for publications by each author.
    
    For each author in authors_json PubMed is queried for the publications. The list of publications is then filtered 
    by prev_pubs, affiliations, and cutoff_year. If the publication is in the of prev_pubs then it is skipped. 
    If the author doesn't have at least one matching affiliation then the publication is skipped. If the publication 
    was published before the cutoff_year then it is skipped. Each publication is then determined to have citations 
    for any of the grants in the author's grants.
    
    Args:
        prev_pubs (dict): dictionary of publications matching the JSON schema for publications.
        authors_json (dict): keys are authors and values are author attributes. Matches Authors section of configuration JSON schema.
        from_email (str): used in the query to PubMed
        
    Returns:
        publication_dict (dict): keys are pulication ids and values are a dictionary with publication attributes
    """
    
    ## Some helpful code to get the xml back as text. import xml.etree.ElementTree as ET    ET.tostring(Element)    
    # initiate PubMed API
    pubmed = pymed.PubMed(tool=TOOL, email=from_email)
    
    publication_dict = dict()
    prev_pubs_titles = [pub_attr["title"] for pub_attr in prev_pubs.values()]
    titles = []
    
    ########################
    # loop through list of authors and request a list of their publications
    ########################
    for author, author_attributes in authors_json.items():
        
        publications = pubmed.query(author_attributes["pubmed_name_search"], max_results=500)
        
        ## Unpacking pub from publications appears to be the slowest part of the code.
        ## publications is an iterator that is broken up into batches and there are noticeable slow downs each time a new batch is fetched.
        for pub in publications:
            
            pub_id = DOI_URL + pub.doi if pub.doi else pub.pubmed_id.split("\n")[0]
            publication_date = int(str(pub.publication_date)[:4])
            
            ## if the publication id is in prev_pubs or publication_dict or the publication date is before the curoff year then skip.
            if publication_date < author_attributes["cutoff_year"] or \
               helper_functions.is_pub_in_publication_dict(pub_id, pub.title, prev_pubs, prev_pubs_titles) or \
               helper_functions.is_pub_in_publication_dict(pub_id, pub.title, publication_dict, titles):
                continue
            
            author_list = helper_functions.match_authors_in_pub_PubMed(authors_json, pub.authors)
    
            ## If no authors were matched then go to the next publication. Note that this is not uncommon because PubMed returns publications for authors who were just colloborators.
            if not author_list:
                continue
                
            pub.authors = author_list
            pub_dict = helper_functions.modify_pub_dict_for_saving(pub)
            
            publication_dict[pub_id] = pub_dict
            titles.append(pub_dict["title"])
                    
            
        # don't piss off NCBI
        time.sleep(1)
        
    return publication_dict


       
        
        
def search_ORCID_for_pubs(prev_pubs, ORCID_key, ORCID_secret, authors_json):
    """Searhes ORCID for publications by each author.
    
    For each author in authors_json ORCID is queried for the publications. The list of publications is then filtered 
    by prev_pubs, affiliations, and cutoff_year. If the publication is in the prev_pubs then it is skipped. 
    If the author doesn't have at least one matching affiliation then the publication is skipped. If the publication 
    was published before the cutoff_year then it is skipped. Each publication is then determined to have citations 
    for any of the grants in the author's grants.
    
    Args:
        prev_pubs (dict): dictionary of publications matching the JSON schema for publications.
        ORCID_key (str): string of the app key ORCID gives when you register the app with them
        ORCID_secret (str): string of the secret ORCID gives when you register the app with them
        authors_json (dict): keys are authors and values are author attributes. Matches authors JSON schema.
        
    Returns:
        publication_dict (dict): keys are publication ids and values are a dictionary with publication attributes
    """
    
    api = orcid.PublicAPI(ORCID_key, ORCID_secret)
    search_token = api.get_search_token_from_orcid()
    
    publication_dict = {}
    prev_pubs_titles = [pub_attr["title"] for pub_attr in prev_pubs.values()]
    titles = []

    for author, authors_attributes in authors_json.items():
        
        if not "ORCID" in authors_attributes:
            continue
        
        summary = api.read_record_public(authors_attributes["ORCID"], 'works',
                                     search_token)
        
        for work in summary["group"]:
            title = None
            doi = None
            external_url = None
            publication_year = None
            publication_month = None
            publication_day = None
            pmid = None
            ## If the work is not a journal article then skip it.
            work_is_a_journal_article = True
            work_before_relevant_year = False
            for work_summary in work["work-summary"]:
                
                if work_summary["type"] != "JOURNAL_ARTICLE":
                    work_is_a_journal_article = False
                    break
                
                
                if work_summary["publication-date"]:
                    if not publication_year and work_summary["publication-date"]["year"]:
                        publication_year = int(work_summary["publication-date"]["year"]["value"])
                        
                    if not publication_month and work_summary["publication-date"]["month"]:
                        publication_month = int(work_summary["publication-date"]["month"]["value"])
                        
                    if not publication_day and work_summary["publication-date"]["day"]:
                        publication_day = int(work_summary["publication-date"]["day"]["value"])

                    if publication_year < authors_attributes["cutoff_year"]:
                        work_before_relevant_year = True
                        break
                
                
                if work_summary["title"] and not title:
                    title = work_summary["title"]["title"]["value"]
                
                if not doi:
                    for external_id in work_summary["external-ids"]["external-id"]:
                        if external_id["external-id-type"] == "doi":
                            doi = external_id["external-id-value"].lower()
                            break
                        elif external_id["external-id-url"]:
                            external_url = external_id["external-id-url"]["value"]
                        elif external_id["external-id-type"] == "pmid":
                            pmid = external_id["external-id-value"]
                
                if title and doi and publication_year and publication_month and publication_day:
                    break
            
            if work_before_relevant_year or not work_is_a_journal_article:
                continue
                       
            if (not title and not doi) or (not doi and not external_url) or not publication_year:
                continue
            else:
                if doi:
                    pub_id = DOI_URL + doi
                elif external_url:
                    pub_id = external_url
                elif pmid:
                    pub_id = pmid
                else:
                    helper_functions.vprint("Warning: Could not find a DOI, URL, or PMID for a publication when searching ORCID. It will not be in the publications", verbosity=1)
                    helper_functions.vprint("Title: " + title, verbosity=1)
                    continue
                    
                if helper_functions.is_pub_in_publication_dict(pub_id, title, prev_pubs, prev_pubs_titles):
                    continue
                    
                
                pub_dict = copy.deepcopy(PUBLICATION_TEMPLATE)
                if doi:
                    pub_dict["doi"] = doi
                if title:
                    pub_dict["title"] = title
                if publication_year:
                    pub_dict["publication_date"]["year"] = publication_year
                if publication_month:
                    pub_dict["publication_date"]["month"] = publication_month
                if publication_day:
                    pub_dict["publication_date"]["day"] = publication_day
                if pmid:
                    pub_dict["pubmed_id"] = pmid
                    
                
                if not helper_functions.is_pub_in_publication_dict(pub_id, title, publication_dict, titles):
                    pub_dict["authors"] = [{"affiliation": ",".join(authors_attributes["affiliations"]),
                                            "firstname": authors_attributes["first_name"],
                                            "initials": None,
                                            "lastname": authors_attributes["last_name"],
                                            "author_id" : author}]
                    publication_dict[pub_id] = pub_dict
                    titles.append(title)
                elif pub_id in publication_dict:
                    author_ids = [pub_author["author_id"] for pub_author in publication_dict[pub_id]["authors"]]
                    if not author in author_ids:
                        publication_dict[pub_id]["authors"].append({"affiliation": ",".join(authors_attributes["affiliations"]),
                                                                     "firstname": authors_attributes["first_name"],
                                                                     "initials": None,
                                                                     "lastname": authors_attributes["last_name"],
                                                                     "author_id" : author})
                
        time.sleep(1)
        
    return publication_dict


            

def search_Google_Scholar_for_pubs(prev_pubs, authors_json, mailto_email):
    """Searhes Google Scholar for publications by each author.
    
    For each author in authors_json Google Scholar is queried for the publications. The list of publications is then filtered 
    by prev_pubs, affiliations, and cutoff_year. If the publication is in the prev_pubs then it is skipped. 
    If the author doesn't have at least one matching affiliation then the publication is skipped. If the publication 
    was published before the cutoff_year then it is skipped. Each publication is then determined to have citations 
    for any of the grants in the author's grants.
    
    Args:
        prev_pubs (dict): dictionary of publications matching the JSON schema for publications.
        authors_json (dict): keys are authors and values are author attributes. Matches authors JSON schema.
        mailto_email (str): used in the query to Crossref when trying to find DOIs for the articles
        
    Returns:
        publication_dict (dict): keys are pulication ids and values are a dictionary with publication attributes
    """
    
    publication_dict = {}
    prev_pubs_titles = [pub_attr["title"] for pub_attr in prev_pubs.values()]
    titles = []
    for author, authors_attributes in authors_json.items():
        
        if not "scholar_id" in authors_attributes:
            continue
        
        try:
            queried_author = scholarly.scholarly.search_author_id(authors_attributes["scholar_id"])
        except:
            helper_functions.vprint("Warning: The \"scholar_id\" for author " + author + " is probably incorrect, an error occured when trying to query Google Scholar.", verbosity=1)    
            continue
        
        if not queried_author["scholar_id"] == authors_attributes["scholar_id"]:
            continue
        
        ## Note that fill modifies the passed dictionary directly, but this is easier to mock in unit tests.
        queried_author = scholarly.scholarly.fill(queried_author, sections=["publications"])
        
        for pub in queried_author["publications"]:
            
            ## Find the publication year and check that it is in range.
            publication_year = int(pub["bib"]["pub_year"]) if "pub_year" in pub["bib"] else None
        
            if not publication_year or publication_year < authors_attributes["cutoff_year"]:
                continue
            
            title = pub["bib"]["title"]
            
            ## Determine the pub_id
            doi = webio.get_DOI_from_Crossref(title, mailto_email)
            if doi:
                pub_id = DOI_URL + doi
            else:
                pub = scholarly.scholarly.fill(pub)
                if "pub_url" in pub:
                    pub_id = pub["pub_url"]
                else:
                    helper_functions.vprint("Warning: Could not find a DOI, URL, or PMID for a publication when searching Google Scholar. It will not be in the publications.", verbosity=1)
                    helper_functions.vprint("Title: " + title, verbosity=1)
                    continue
            
            if helper_functions.is_pub_in_publication_dict(pub_id, title, prev_pubs, prev_pubs_titles):
                continue
        
            
            pub_dict = copy.deepcopy(PUBLICATION_TEMPLATE)
            if doi:
                pub_dict["doi"] = doi
            if title:
                pub_dict["title"] = title
            if publication_year:
                pub_dict["publication_date"]["year"] = publication_year
                
            
            if not helper_functions.is_pub_in_publication_dict(pub_id, title, publication_dict, titles):
                pub_dict["authors"] = [{"affiliation": ",".join(authors_attributes["affiliations"]),
                                        "firstname": authors_attributes["first_name"],
                                        "initials": None,
                                        "lastname": authors_attributes["last_name"],
                                        "author_id" : author}]
                publication_dict[pub_id] = pub_dict
                titles.append(title)
            elif pub_id in publication_dict:
                author_ids = [pub_author["author_id"] for pub_author in publication_dict[pub_id]["authors"]]
                if not author in author_ids:
                    publication_dict[pub_id]["authors"].append({"affiliation": ",".join(authors_attributes["affiliations"]),
                                             "firstname": authors_attributes["first_name"],
                                             "initials": None,
                                             "lastname": authors_attributes["last_name"],
                                             "author_id" : author})
                
        time.sleep(1)
            
    return publication_dict





def search_Crossref_for_pubs(prev_pubs, authors_json, mailto_email):
    """Searhes Crossref for publications by each author.
    
    For each author in authors_json Crossref is queried for the publications. The list of publications is then filtered 
    by prev_pubs, affiliations, and cutoff_year. If the publication is in the prev_pubs then it is skipped. 
    If the author doesn't have at least one matching affiliation then the publication is skipped. If the publication 
    was published before the cutoff_year then it is skipped. Each publication is then determined to have citations 
    for any of the grants in the author's grants.
    
    Args:
        prev_pubs (dict): dictionary of publications matching the JSON schema for publications.
        authors_json (dict): keys are authors and values are author attributes. Matches authors JSON schema.
        mailto_email (str): used in the query to Crossref
        
    Returns:
        publication_dict (dict): keys are pulication ids and values are a dictionary with publication attributes
    """
    
    cr = habanero.Crossref(ua_string = "Academic Tracker (mailto:" + mailto_email + ")")
    
    publication_dict = {}
    prev_pubs_titles = [pub_attr["title"] for pub_attr in prev_pubs.values()]
    titles = []
    for author, authors_attributes in authors_json.items():
        
        results = cr.works(query_author = authors_attributes["pubmed_name_search"], filter = {"type":"journal-article", "from-pub-date":str(authors_attributes["cutoff_year"])}, limit = 300)
        
        for work in results["message"]["items"]:
            
            ## Find published date
            date_found = False
            if "published" in work:
                date_key = "published"
                date_found = True
            elif "published-online" in work:
                date_key = "published-online"
                date_found = True
            elif "published-print" in work:
                date_key = "published-print"
                date_found = True
            
            if date_found:
                date_length = len(work[date_key]["date-parts"])
                
                if date_length > 2:
                    publication_year = work[date_key]["date-parts"][0][0]
                    publication_month = work[date_key]["date-parts"][0][1]
                    publication_day = work[date_key]["date-parts"][0][2]
                elif date_length > 1:
                    publication_year = work[date_key]["date-parts"][0][0]
                    publication_month = work[date_key]["date-parts"][0][1]
                    publication_day = None
                elif date_length > 0:
                    publication_year = work[date_key]["date-parts"][0][0]
                    publication_month = None
                    publication_day = None
                else:
                    publication_year = None
                    publication_month = None
                    publication_day = None
            else:
                publication_year = None
                publication_month = None
                publication_day = None
            
            
                
            if publication_year < authors_attributes["cutoff_year"]:
                continue
            
            if "title" in work:
                title = work["title"][0]
            else:
                continue
            
            author_list = helper_functions.match_authors_in_pub_Crossref(authors_json, work["author"])
            ## If the author_list is empty then there were no matching authors, continue.
            if not author_list:
                continue
            
            ## Change the author list to a form like PubMed's
            new_author_list = []
            for cr_author_dict in author_list:
                
                temp_dict = {"lastname":cr_author_dict["family"], "initials":None,}
                
                temp_dict["firstname"] = cr_author_dict["given"] if "given" in cr_author_dict else None
                
                if cr_author_dict["affiliation"] and "name" in cr_author_dict["affiliation"][0]:
                    temp_dict["affiliation"] = cr_author_dict["affiliation"][0]["name"]
                else:
                    temp_dict["affiliation"] = None
                    
                if "author_id" in cr_author_dict:
                    temp_dict["author_id"] = cr_author_dict["author_id"]
                
                new_author_list.append(temp_dict)
            
            ## Look for DOI
            doi = work["DOI"].lower() if "DOI" in work else None
            
            ## Look for external URL
            if "URL" in work:
                external_url = work["URL"]
            elif "link" in work:
                external_url = [link["URL"] for link in work["link"] if "URL" in link and link["URL"]][0]
                if not external_url:
                    external_url = None
            else:
                external_url = None
                            
            
            if doi:
                pub_id = DOI_URL + doi
            elif external_url:
                pub_id = external_url
            else:
                helper_functions.vprint("Warning: Could not find a DOI or external URL for a publication when searching Crossref. It will not be in the publications.", verbosity=1)
                helper_functions.vprint("Title: " + title, verbosity=1)
                continue
            
            
            
            if helper_functions.is_pub_in_publication_dict(pub_id, title, prev_pubs, prev_pubs_titles) or \
               helper_functions.is_pub_in_publication_dict(pub_id, title, publication_dict, titles):
                continue
            
            
            
            ## look for grants in results
            if "funder" in work:
                ## the grant string could be in any value of the funder dict so look for it in each one.
                found_grants = {grant for funder in work["funder"] for value in funder.values() for grant in authors_attributes["grants"] if grant in value}
                found_grants = list(found_grants) if found_grants else None
            else:
                found_grants = None
                
            ## Look for journal
            journal = work["publisher"] if "publisher" in work else None
                
        
            ## Build the pub_dict from what we were able to collect.
            pub_dict = copy.deepcopy(PUBLICATION_TEMPLATE)
            if doi:
                pub_dict["doi"] = doi
            if publication_year:
                pub_dict["publication_date"]["year"] = publication_year
            if publication_month:
                pub_dict["publication_date"]["month"] = publication_month
            if publication_day:
                pub_dict["publication_date"]["day"] = publication_day
            if journal:
                pub_dict["journal"] = journal
            if found_grants:
                pub_dict["grants"] = found_grants
                
            pub_dict["authors"] = new_author_list
            pub_dict["title"] = title
            
            
            publication_dict[pub_id] = pub_dict
            titles.append(title)
            
        time.sleep(1)
            
            
    return publication_dict



# -*- coding: utf-8 -*-
"""
Reference Search Webio
~~~~~~~~~~~~~~~~~~~~~~

Internet interfacing for reference_search.
"""

import time
import copy
import sys
import os
import re

import pymed
import scholarly
import habanero
import bs4
import fuzzywuzzy.fuzz

from . import helper_functions
from . import citation_parsing
from . import webio
from . import fileio
from . import user_input_checking


TOOL = webio.TOOL
DOI_URL = webio.DOI_URL

PUBLICATION_TEMPLATE = webio.PUBLICATION_TEMPLATE


def build_pub_dict_from_PMID(PMID_list, from_email):
    """Query PubMed for each PMID and build a dictionary of the returned data.
    
    Args:
        PMID_list (list): A list of PMIDs as strings.
        from_email (str): An email address to use when querying PubMed.
        
    Returns:
        publication_dict (dict): keys are pulication ids and values are a dictionary with publication attributes.
    """
    
    pubmed = pymed.PubMed(tool=TOOL, email=from_email)
    
    publication_dict = dict()
    
    for PMID_to_search in PMID_list:
        
        publications = pubmed.query(PMID_to_search, max_results=10)
        
        for pub in publications:
            
            pmid = pub.pubmed_id.split("\n")[0]
            pub_id = DOI_URL + pub.doi.lower() if pub.doi else pmid
            if pmid == PMID_to_search:
                publication_dict[pub_id] = helper_functions.modify_pub_dict_for_saving(pub)
                break
                
        time.sleep(1)

    return publication_dict



def search_references_on_PubMed(tokenized_citations, from_email):
    """Searhes PubMed for publications matching the citations.
    
    For each citation in tokenized_citations PubMed is queried for the publication. 
    
    Args:
        tokenized_citations (list): list of citations parsed from a source. Each citation is a dict {"authors", "title", "DOI", "PMID", "reference_line", "pub_dict_key"}.
        from_email (str): used in the query to PubMed
        
    Returns:
        publication_dict (dict): keys are pulication ids and values are a dictionary with publication attributes
    """
       
    # initiate PubMed API
    pubmed = pymed.PubMed(tool=TOOL, email=from_email)
    
    publication_dict = dict()
    titles = []
    matching_key_for_citation = []
    
    for citation in tokenized_citations:
        
        if citation["PMID"]:
            query_string = citation["PMID"]
        elif citation["DOI"]:
            query_string = citation["DOI"]
        elif citation["title"]:
            query_string = citation["title"]
        else:
            matching_key_for_citation.append(None)
            continue
        
        publications = pubmed.query(query_string, max_results=10)
        
        citation_matched = False
        for pub in publications:
            
            pmid = pub.pubmed_id.split("\n")[0]
            pub_id = DOI_URL + pub.doi.lower() if pub.doi else pmid
            
            if helper_functions.is_pub_in_publication_dict(pub_id, pub.title, publication_dict, titles):
                continue
            
            ## Match publication to the citation.
            pub_matched = False
            if citation["PMID"] == pmid:
                pub_matched = True
            elif pub.doi and citation["DOI"] == pub.doi.lower():
                pub_matched = True
            else:
                has_matching_author = any([author_items.get("lastname").lower() == author_attributes["last"].lower() for author_items in pub.authors for author_attributes in citation["authors"] if author_items.get("lastname") and author_attributes["last"]])
                if has_matching_author and fuzzywuzzy.fuzz.ratio(citation["title"], pub.title) >= 90:
                    pub_matched = True
                               
            if not pub_matched:
                continue
                        
            
            pub_dict = helper_functions.modify_pub_dict_for_saving(pub)
            
            publication_dict[pub_id] = pub_dict
            titles.append(pub_dict["title"])
            matching_key_for_citation.append(pub_id)
            citation_matched = True
            break
                    
        if not citation_matched:
            matching_key_for_citation.append(None)
        time.sleep(1)
        
    return publication_dict, matching_key_for_citation


       

def search_references_on_Google_Scholar(tokenized_citations, mailto_email):
    """Searhes Google Scholar for publications that match the citations.
        
    Args:
        tokenized_citations (list): list of citations parsed from a source. Each citation is a dict {"authors", "title", "DOI", "PMID", "reference_line", "pub_dict_key"}.
        mailto_email (str): used in the query to Crossref when trying to find DOIs for the articles
        
    Returns:
        publication_dict (dict): keys are pulication ids and values are a dictionary with publication attributes
    """
    
    publication_dict = {}
    titles = []
    matching_key_for_citation = []
    for citation in tokenized_citations:
        
        if citation["title"]:
            query = scholarly.scholarly.search_pubs(citation["title"])
        else:
            matching_key_for_citation.append(None)
            continue
        
        citation_matched = False
        for count, pub in enumerate(query):
            
            if count > 50:
                break
            
            time.sleep(1)
            
            title = pub["bib"]["title"]
            
            ## authors from Google Scholar are last names and initials in a single string, each string in one list. ['SA Cholewiak', 'RW Fleming', 'M Singh']
            pub_matched = False
            has_matching_author = any([author_attributes["last"].lower() in author.lower() for author in pub["bib"]["author"] for author_attributes in citation["authors"]])
            if has_matching_author and fuzzywuzzy.fuzz.ratio(citation["title"], title) >= 90:
                pub_matched = True
                               
            if not pub_matched:
                continue
            
            ## Find the publication year and check that it is in range.
            publication_year = int(pub["bib"]["pub_year"]) if "pub_year" in pub["bib"] else None
        
            
            ## Determine the pub_id
            doi = webio.get_DOI_from_Crossref(title, mailto_email)
            if doi:
                pub_id = DOI_URL + doi
            else:
                if "pub_url" in pub:
                    pub_id = pub["pub_url"]
                else:
                    helper_functions.vprint("Warning: Could not find a DOI, URL, or PMID for a publication when searching Google Scholar. It will not be in the publications.", verbosity=1)
                    helper_functions.vprint("Title: " + title, verbosity=1)
                    break
            
            
            pub_dict = copy.deepcopy(PUBLICATION_TEMPLATE)
            if doi:
                pub_dict["doi"] = doi
            if title:
                pub_dict["title"] = title
            if publication_year:
                pub_dict["publication_date"]["year"] = publication_year
                
            
            if not helper_functions.is_pub_in_publication_dict(pub_id, title, publication_dict, titles):
                pub_dict["authors"] = [{"affiliation": None,
                                        "firstname": None,
                                        "initials": None,
                                        "lastname": author["last"]} for author in citation["authors"]]
                publication_dict[pub_id] = pub_dict
                titles.append(title)
                matching_key_for_citation.append(pub_id)
                citation_matched = True
            
            break
        
        if not citation_matched:
            matching_key_for_citation.append(None)
        
    return publication_dict, matching_key_for_citation




def search_references_on_Crossref(tokenized_citations, mailto_email):
    """Searhes Crossref for publications matching the citations.
    
    Args:
        tokenized_citations (list): list of citations parsed from a source. Each citation is a dict {"authors", "title", "DOI", "PMID", "reference_line", "pub_dict_key"}.
        mailto_email (str): used in the query to Crossref
        
    Returns:
        publication_dict (dict): keys are pulication ids and values are a dictionary with publication attributes
    """
    
    cr = habanero.Crossref(ua_string = "Academic Tracker (mailto:" + mailto_email + ")")
    
    publication_dict = {}
    titles = []
    matching_key_for_citation = []
    for citation in tokenized_citations:
        
        if citation["DOI"]:
            results = cr.works(ids = citation["DOI"])
            works = [results["message"]]
        elif citation["title"]:
            results = cr.works(query_bibliographic = citation["title"], filter = {"type":"journal-article"}, limit = 10)
            works = results["message"]["items"]
        else:
            matching_key_for_citation.append(None)
            continue
        
        citation_matched = False
        for work in works:
            
            ## Look for DOI
            doi = work["DOI"].lower() if "DOI" in work else None
             
            ## Look for title    
            if "title" in work and work["title"]:
                title = work["title"][0]
            else:
                continue
            
            pub_matched = False
            if citation["DOI"] == doi:
                pub_matched = True
            else:
                if "author" in work:
                    has_matching_author = any([author_items.get("family").lower() == author_attributes["last"].lower() for author_items in work["author"] for author_attributes in citation["authors"] if "family" in author_items and author_items.get("family")])
                    if has_matching_author and fuzzywuzzy.fuzz.ratio(citation["title"], title) >= 90:
                        pub_matched = True
                               
            if not pub_matched:
                continue
            
            
            ## Find published date
            date_found = False
            if "published" in work:
                date_key = "published"
                date_found = True
            elif "published-online" in work:
                date_key = "published-online"
                date_found = True
            elif "published-print" in work:
                date_key = "published-print"
                date_found = True
            
            if date_found:
                date_length = len(work[date_key]["date-parts"])
                
                if date_length > 2:
                    publication_year = work[date_key]["date-parts"][0][0]
                    publication_month = work[date_key]["date-parts"][0][1]
                    publication_day = work[date_key]["date-parts"][0][2]
                elif date_length > 1:
                    publication_year = work[date_key]["date-parts"][0][0]
                    publication_month = work[date_key]["date-parts"][0][1]
                    publication_day = None
                elif date_length > 0:
                    publication_year = work[date_key]["date-parts"][0][0]
                    publication_month = None
                    publication_day = None
                else:
                    publication_year = None
                    publication_month = None
                    publication_day = None
            else:
                publication_year = None
                publication_month = None
                publication_day = None
            
            
            ## Change the author list to a form like PubMed's
            new_author_list = []
            for cr_author_dict in work["author"]:
                
                temp_dict = {"lastname":cr_author_dict["family"], "initials":None,}
                
                temp_dict["firstname"] = cr_author_dict["given"] if "given" in cr_author_dict else None
                
                if cr_author_dict["affiliation"] and "name" in cr_author_dict["affiliation"][0]:
                    temp_dict["affiliation"] = cr_author_dict["affiliation"][0]["name"]
                else:
                    temp_dict["affiliation"] = None
                    
                if "author_id" in cr_author_dict:
                    temp_dict["author_id"] = cr_author_dict["author_id"]
                
                new_author_list.append(temp_dict)
            
            
            ## Look for external URL
            if "URL" in work:
                external_url = work["URL"]
            elif "link" in work:
                external_url = [link["URL"] for link in work["link"] if "URL" in link and link["URL"]][0]
                if not external_url:
                    external_url = None
            else:
                external_url = None
                            
            
            if doi:
                pub_id = DOI_URL + doi
            elif external_url:
                pub_id = external_url
            else:
                helper_functions.vprint("Warning: Could not find a DOI or external URL for a publication when searching Crossref. It will not be in the publications.", verbosity=1)
                helper_functions.vprint("Title: " + title, verbosity=1)
                continue
            
            
            
            if helper_functions.is_pub_in_publication_dict(pub_id, title, publication_dict, titles):
                continue
            
            
            ## Look for journal
            journal = work["publisher"] if "publisher" in work else None
                
        
            ## Build the pub_dict from what we were able to collect.
            pub_dict = copy.deepcopy(PUBLICATION_TEMPLATE)
            if doi:
                pub_dict["doi"] = doi
            if publication_year:
                pub_dict["publication_date"]["year"] = publication_year
            if publication_month:
                pub_dict["publication_date"]["month"] = publication_month
            if publication_day:
                pub_dict["publication_date"]["day"] = publication_day
            if journal:
                pub_dict["journal"] = journal
                
            pub_dict["authors"] = new_author_list
            pub_dict["title"] = title
            
            
            publication_dict[pub_id] = pub_dict
            titles.append(title)
            matching_key_for_citation.append(pub_id)
            citation_matched = True
            break
        
        if not citation_matched:
            matching_key_for_citation.append(None)
        time.sleep(1)
            
            
    return publication_dict, matching_key_for_citation
        
                



def parse_myncbi_citations(url):
    """Tokenize the citations on a MyNCBI URL.
    
    Note that authors and title can be missing or empty from the webpage.
    This function assumes the url is the first page of the MyNCBI citations.
    The first page is tokenized and then each subsequent page is visited and 
    tokenized.
    
    Args:
        url (str): the url of the MyNCBI page.
        
    Returns:
        parsed_pubs (dict): the citations tokenized in a dictionary matching the tokenized citations JSON schema.    
    """
    
    ## Get the first page, find out the total pages, and parse it.
    url_str = webio.get_url_contents_as_str(url)
    if not url_str:
        helper_functions.vprint("Error: Could not access the MYNCBI webpage. Make sure the address is correct.")
        sys.exit()
    
    soup = bs4.BeautifulSoup(url_str, "html.parser")
    number_of_pages = int(soup.find("span", class_ = "totalPages").text)
    
    parsed_pubs = citation_parsing.tokenize_myncbi_citations(url_str)
    
    ## Parse the rest of the pages.
    new_url = url if url[-1] == "/" else url + "/"
    
    for i in range(2,number_of_pages+1):
        
        url_str = webio.get_url_contents_as_str(new_url + "?page=" + str(i))
        if not url_str:
            helper_functions.vprint("Error: Could not access page " + str(i) + " of the MYNCBI webpage. Aborting run.")
            sys.exit()
        
        temp_pubs = citation_parsing.tokenize_myncbi_citations(url_str)
        parsed_pubs += temp_pubs
        
    return parsed_pubs



def tokenize_reference_input(reference_input, MEDLINE_reference):
    """Tokenize the citations in reference_input.
    
    reference_input can be a URL or filepath. MyNCBI URLs are handled special, 
    but all other URLs are read as a text document and parsed line by line as 
    if they were a test document. If the format of the reference is MEDLINE then 
    set MEDLINE_reference to True and it will be parsed as such instead of line 
    by line. Citations are expected to be 1 per line otherwise.
    
    Args:
        reference_input (str): URL or filepath
        MEDLINE_reference (bool): True if reference_input is in MEDLINE format
        
    Returns:
        tokenized_citations (dict): the citations tokenized in a dictionary matching the tokenized citations JSON schema. 
    """
    
    ## Check the reference_input to see if it is json.
    extension = os.path.splitext(reference_input)[1][1:]
    if extension == "json":
        tokenized_citations = fileio.load_json(reference_input)
        user_input_checking.tok_reference_check(tokenized_citations)
    
    ## Check the reference file input and see if it is a URL
    elif re.match(r"http.*", reference_input):
        if re.match(r".*ncbi.nlm.nih.gov/myncbi.*", reference_input):
            tokenized_citations = parse_myncbi_citations(reference_input)
        else:
            document_string = webio.clean_tags_from_url(reference_input)
            
            if not document_string:
                helper_functions.vprint("Nothing was read from the URL. Make sure the address is correct.")
                sys.exit()
            
            tokenized_citations = citation_parsing.parse_text_for_citations(document_string)
    else:
        # Check the file extension and call the correct read in function.
        if extension == "docx":
            document_string = fileio.read_text_from_docx(reference_input)
        elif extension == "txt":
            document_string = fileio.read_text_from_txt(reference_input)
        else:
            helper_functions.vprint("Unknown file type for reference file.")
            sys.exit()
    
        if not document_string:
            helper_functions.vprint("Nothing was read from the reference file. Make sure the file is not empty or is a supported file type.")
            sys.exit()
        
        if MEDLINE_reference:
            tokenized_citations = citation_parsing.parse_MEDLINE_format(document_string)
        else:
            tokenized_citations = citation_parsing.parse_text_for_citations(document_string)
            
    if not tokenized_citations:
        helper_functions.vprint("Warning: Could not tokenize any citations in provided reference. Check setup and formatting and try again.")
        sys.exit()
    
    ## Look for duplicates in citations.
    duplicate_citations = helper_functions.find_duplicate_citations(tokenized_citations)
    if duplicate_citations:
        helper_functions.vprint("Warning: The following citations in the reference file or URL appear to be duplicates based on identical DOI, PMID, or similar titles. They will only appear once in any outputs.", verbosity=1)
        helper_functions.vprint("Duplicates:", verbosity=1)
        for index_list in duplicate_citations:
            for index in index_list:
                if tokenized_citations[index]["reference_line"]:
                    pretty_print = tokenized_citations[index]["reference_line"].split("\n")
                    pretty_print = " ".join([line.strip() for line in pretty_print])
                    helper_functions.vprint(pretty_print, verbosity=1)
                else:
                    helper_functions.vprint(tokenized_citations[index]["title"], verbosity=1)
                helper_functions.vprint("", verbosity=1)
            helper_functions.vprint("\n", verbosity=1)
        
        indexes_to_remove = [index for duplicate_set in duplicate_citations for index in duplicate_set[1:]]
        
        tokenized_citations = [citation for count, citation in enumerate(tokenized_citations) if not count in indexes_to_remove]
        
    return tokenized_citations









# -*- coding: utf-8 -*-
"""
Reference Search Modularized
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Modularized pieces of reference_search.
"""

import re
import datetime
import os

from . import user_input_checking
from . import fileio
from . import helper_functions
from . import ref_srch_webio
from . import ref_srch_emails_and_reports
from . import webio



def input_reading_and_checking(config_json_filepath, ref_path_or_URL, MEDLINE_reference, no_Crossref, prev_pub_filepath):
    """Read in inputs from user and do error checking.
    
    Args:
        config_json_filepath (str): filepath to the configuration JSON.
        ref_path_or_URL (str): either a filepath to file to tokenize or a URL to tokenize.
        MEDLINE_reference (bool): If True re_path_or_URL is a filepath to a MEDLINE formatted file.
        no_Crossref (bool): If True search Crossref else don't. Reduces checking on config JSON if True.
        prev_pub_filepath (str or None): filepath to the publication JSON to read in.
        
    Returns:
        config_dict (dict): Matches the Configuration file JSON schema.
        tokenized_citations (list): list of dicts. Matches the tokenized citations JSON schema.
        has_previous_pubs (bool): True if a prev_pub file was input, False otherwise.
        prev_pubs (dict): The contents of the prev_pub file input by the user if provided.
    """
    
    ## read in config file
    config_dict = fileio.load_json(config_json_filepath)
    
    ## Get inputs from config file and check them for errors.
    user_input_checking.ref_config_file_check(config_dict, no_Crossref)
    
    if not prev_pub_filepath or prev_pub_filepath.lower() == "ignore":
        prev_pubs = {}
        has_previous_pubs = False
    else:
        prev_pubs = fileio.load_json(prev_pub_filepath)
        has_previous_pubs = True
    
    if has_previous_pubs:
        user_input_checking.prev_pubs_file_check(prev_pubs)
        
    tokenized_citations = ref_srch_webio.tokenize_reference_input(ref_path_or_URL, MEDLINE_reference) 
    
    return config_dict, tokenized_citations, has_previous_pubs, prev_pubs



def build_publication_dict(config_dict, tokenized_citations, no_Crossref):
    """Query PubMed and Crossref for publications matching the citations in tokenized_citations.
    
    Args:
        config_dict (dict): Matches the Configuration file JSON schema.
        tokenized_citations (list): list of dicts. Matches the tokenized citations JSON schema.
        no_Crossref (bool): If True search Crossref else don't. Reduces checking on config JSON if True.
        
    Returns:
        publication_dict (dict): The dictionary matching the publication JSON schema.
        tokenized_citations (list): Same list as the input but with the pud_dict_key updated to match the publication found.        
    """
    
    helper_functions.vprint("Finding publications. This could take a while.")
    helper_functions.vprint("Searching PubMed.")
    PubMed_publication_dict, PubMed_matching_key_for_citation = ref_srch_webio.search_references_on_PubMed(tokenized_citations, config_dict["PubMed_search"]["PubMed_email"])
#    if not args["--no_GoogleScholar"]:
#        print("Searching Google Scholar.")
#        Google_Scholar_publication_dict, Google_Scholar_matching_key_for_citation = webio.search_references_on_Google_Scholar(tokenized_citations, config_dict["Crossref_search"]["Crossref_email"], args["--verbose"])
    if not no_Crossref:
        helper_functions.vprint("Searching Crossref.")
        Crossref_publication_dict, Crossref_matching_key_for_citation = ref_srch_webio.search_references_on_Crossref(tokenized_citations, config_dict["Crossref_search"]["mailto_email"])
    
    publication_dict = PubMed_publication_dict
#    if not args["--no_GoogleScholar"]:
#        for key, value in Google_Scholar_publication_dict.items():
#            if not key in publication_dict:
#                publication_dict[key] = value
    if not no_Crossref:
        for key, value in Crossref_publication_dict.items():
            if not key in publication_dict:
                publication_dict[key] = value
            
    matching_key_for_citation = PubMed_matching_key_for_citation
#    if not args["--no_GoogleScholar"]:
#        matching_key_for_citation = [key if key else Google_Scholar_matching_key_for_citation[count] for count, key in enumerate(matching_key_for_citation)]
    if not no_Crossref:
        matching_key_for_citation = [key if key else Crossref_matching_key_for_citation[count] for count, key in enumerate(matching_key_for_citation)]
        
    for count, citation in enumerate(tokenized_citations):
        if matching_key_for_citation[count]:
            citation["pub_dict_key"] = matching_key_for_citation[count]
            
    return publication_dict, tokenized_citations



def save_and_send_reports_and_emails(config_dict, tokenized_citations, publication_dict, prev_pubs, has_previous_pubs, test):
    """Build the summary report and email it.
    
    Args:
        config_dict (dict): Matches the Configuration file JSON schema.
        tokenized_citations (list): list of dicts. Matches the tokenized citations JSON schema.
        publication_dict (dict): The dictionary matching the publication JSON schema.
        prev_pubs (dict): The contents of the prev_pub file input by the user if provided.
        has_previous_pubs (bool): True if a prev_pub file was input, False otherwise.
        test (bool): If True save_dir_name is tracker-test instead of tracker- and emails are not sent.
        
    Returns:
        save_dir_name (str): Name of the directory where the emails and report were saved.
    """
    
    ## Compare citations to prev_pubs 
    is_citation_in_prev_pubs_list = []
    if has_previous_pubs:
        is_citation_in_prev_pubs_list = helper_functions.are_citations_in_pub_dict(tokenized_citations, prev_pubs)
        
    
    ## Build the save directory name.
    if test:
        save_dir_name = "tracker-test-" + re.sub(r"\-| |\:", "", str(datetime.datetime.now())[2:16])
    else:
        save_dir_name = "tracker-" + re.sub(r"\-| |\:", "", str(datetime.datetime.now())[2:16])
    os.mkdir(save_dir_name)
    
    
    if "summary_report" in config_dict:
        
        if "columns" in config_dict["summary_report"]:
            summary_report, summary_filename = ref_srch_emails_and_reports.create_tabular_report(publication_dict, config_dict, is_citation_in_prev_pubs_list, tokenized_citations, save_dir_name)
        else:
            if "template" in config_dict["summary_report"]:
                template = config_dict["summary_report"]["template"]
            else:
                template = ref_srch_emails_and_reports.DEFAULT_SUMMARY_TEMPLATE
                
            if "filename" in config_dict["summary_report"]:
                summary_filename = config_dict["summary_report"]["filename"]
            else:
                summary_filename = "summary_report.txt"
            
            summary_report = ref_srch_emails_and_reports.create_report_from_template(publication_dict, is_citation_in_prev_pubs_list, tokenized_citations, template)
            fileio.save_string_to_file(save_dir_name, summary_filename, summary_report)
        
        if "from_email" in config_dict["summary_report"]:
            email_messages = {"creation_date" : str(datetime.datetime.now())[0:16]}
            email_messages["emails"] = [{"to":",".join([email for email in config_dict["summary_report"]["to_email"]]),
                                         "from":config_dict["summary_report"]["from_email"],
                                         "cc":",".join([email for email in config_dict["summary_report"]["cc_email"]]) if "cc_email" in config_dict["summary_report"] else "",
                                         "subject":config_dict["summary_report"]["email_subject"],
                                         "body":config_dict["summary_report"]["email_body"],
                                         "attachment":summary_report,
                                         "attachment_filename": summary_filename}]
            
            fileio.save_emails_to_file(email_messages, save_dir_name)
        
            ## send emails
            if not test:
                webio.send_emails(email_messages)
                
    return save_dir_name




